from __future__ import annotations

import re
from collections.abc import Iterable

import click

from commodore.config import Config
from commodore.helpers import kapitan_inventory

from .tools import format_component_list

PACKAGE_PREFIX: str = "pkg."

RESERVED_PACKAGE_PATTERN = re.compile("^(components|defaults|global|params)$")
TENANT_PREFIX_PATTERN = re.compile("^t-.*$")


def _extract_component_aliases(
    cfg: Config, kapitan_applications: Iterable[str]
) -> tuple[set[str], dict[str, set[str]]]:
    """
    Extract components and all aliases from Kapitan applications array.

    This function doesn't validate the resulting data. Generally, callers will want to
    use `_discover_components()` to extract components and their aliases from the
    applications array.

    This function drops any packages included through the applications array.
    """
    components = set()
    all_component_aliases: dict[str, set[str]] = {}
    for component in kapitan_applications:
        if component.startswith(PACKAGE_PREFIX):
            continue
        try:
            cn, alias = component.split(" as ")
        except ValueError:
            cn = component
            alias = component
        if cfg.debug:
            msg = f"   > Found component {cn}"
            if alias != component:
                msg += f" aliased to {alias}"
            click.echo(msg)
        components.add(cn)
        all_component_aliases.setdefault(alias, set()).add(cn)

    return components, all_component_aliases


def _discover_components(cfg) -> tuple[list[str], dict[str, str]]:
    """
    Discover components used by the current cluster by extracting all entries from the
    reclass applications dictionary.

    The function also verifies the extracted entries, and raises an exception if any
    invalid aliases are found.
    """
    kapitan_applications = kapitan_inventory(cfg, key="applications")

    components, all_component_aliases = _extract_component_aliases(
        cfg, kapitan_applications.keys()
    )

    component_aliases: dict[str, str] = {}

    for alias, cns in all_component_aliases.items():
        if len(cns) == 0:
            # NOTE(sg): This should never happen, but we add it for completeness' sake.
            raise ValueError(
                f"Discovered component alias '{alias}' with no associated components"
            )

        if len(cns) > 1:
            if alias in cns:
                other_aliases = list(cns - set([alias]))
                if len(other_aliases) > 1:
                    clist = format_component_list(other_aliases)
                    raise KeyError(
                        f"Components {clist} alias existing component '{alias}'"
                    )

                # If this assertion fails we have a problem, since `other_aliases` is
                # the result of removing a single element from a set which contains
                # multiple elements and we've already handled the case for len() > 1.
                # Since we don't mind if it's optimized out in some cases, we annotate
                # it with `nosec` so bandit doesn't complain about it.
                assert len(other_aliases) == 1  # nosec
                raise KeyError(
                    f"Component '{other_aliases[0]}' "
                    + f"aliases existing component '{alias}'"
                )

            clist = format_component_list(cns)
            raise KeyError(
                f"Duplicate component alias '{alias}': "
                + f"components {clist} are aliased to '{alias}'"
            )

        # len(cns) must be 1 here, as we already raise an Exception for len(cns) ==
        # 0 earlier. We still assert this condition here and annotate with `nosec`
        # so bandit doesn't complain about it.
        assert len(cns) == 1  # nosec
        component_aliases[alias] = list(cns)[0]

    return sorted(components), component_aliases


def _discover_packages(cfg: Config) -> set[str]:
    """
    Discover configuration packages included through the applications array.

    All config package inclusions must be prefixed with `pkg.`. This function drops any
    component includes. To parse components from the applications array, use
    `_discover_components()`.
    """
    kapitan_applications = kapitan_inventory(
        cfg, key="applications", ignore_class_notfound=True
    )

    packages = set()

    for app in kapitan_applications:
        if not app.startswith(PACKAGE_PREFIX):
            continue

        pkgname = app.replace(PACKAGE_PREFIX, "", 1)

        if RESERVED_PACKAGE_PATTERN.match(pkgname):
            raise click.ClickException(
                f"Can't use reserved name '{pkgname}' as package name."
            )

        if TENANT_PREFIX_PATTERN.match(pkgname):
            raise click.ClickException(
                "Package names can't be prefixed with 't-'."
                + " This prefix is reserved for tenant configurations."
            )

        packages.add(pkgname)

    return packages



from __future__ import annotations

import shutil

from collections.abc import Iterable
from pathlib import Path
from tempfile import mkdtemp
from textwrap import dedent
from typing import Optional

import click

from commodore.compile import clean_working_tree
from commodore.cluster import update_target
from commodore.compile import setup_compile_environment
from commodore.config import Config
from commodore.dependency_mgmt import fetch_components, register_components
from commodore.helpers import kapitan_compile, relsymlink, yaml_dump
from commodore.inventory import Inventory
from commodore.postprocess import postprocess_components


def resolve_and_create_work_dir(
    cfg: Config, tmp_dir: Optional[str], keep_dir: bool
) -> tuple[Path, bool]:
    if tmp_dir:
        if not tmp_dir.startswith("/"):
            temp_dir = Path(cfg.work_dir, tmp_dir).resolve()
        else:
            temp_dir = Path(tmp_dir).resolve()
        if cfg.debug:
            click.echo(
                " > Always setting `--keep-dir` when temp dir provided explicitly"
            )
        keep_dir = True
    else:
        temp_dir = Path(mkdtemp(prefix="package-")).resolve()

    return temp_dir, keep_dir


# pylint: disable=too-many-arguments disable=too-many-locals
def compile_package(
    cfg: Config,
    pkg_path_: str,
    root_class: str,
    value_files_: Iterable[str],
    tmp_dir: Optional[str] = "",
    keep_dir: bool = False,
):
    temp_dir, keep_dir = resolve_and_create_work_dir(cfg, tmp_dir, keep_dir)
    cfg.work_dir = temp_dir

    # Clean working tree before compiling package, some of our symlinking logic expects
    # that `inventory/` is cleaned before symlinks are created.
    if not cfg.local:
        clean_working_tree(cfg)

    pkg_path = Path(pkg_path_).resolve()
    value_files = [Path(f).resolve() for f in value_files_]

    # NOTE(sg): The assumption here is that package repo names generally reflect the
    # package name (possibly with a prefix `package-`). As long as a package only uses
    # relative includes internally, the package name shouldn't matter for compilation.
    pkg_name = pkg_path.stem.replace("package-", "")
    # Handle package directory naming scheme for packages which are part of a cluster
    # catalog separately.
    if pkg_path.name.startswith("pkg."):
        pkg_name = pkg_path.name.replace("pkg.", "")

    root_class_name = root_class.replace(".yml", "").replace("/", ".")
    # Convert root class file to class name
    root_class_path = pkg_path / root_class
    if not root_class_path.is_file():
        raise click.ClickException(f"Test class '{root_class_name}' doesn't exist.")

    _setup_inventory(cfg.inventory, pkg_name, root_class_name, value_files)

    # Symlink package to inventory
    relsymlink(pkg_path, cfg.inventory.package_dir(pkg_name).parent, pkg_name)

    # setup bootstrap target
    update_target(cfg, cfg.inventory.bootstrap_target)

    if not cfg.local:
        fetch_components(cfg)
    else:
        register_components(cfg)

    update_target(cfg, cfg.inventory.bootstrap_target)

    aliases = cfg.get_component_aliases()
    for alias, c in aliases.items():
        update_target(cfg, alias, c)

    inventory, targets = setup_compile_environment(cfg)

    kapitan_compile(cfg, targets, search_paths=[cfg.vendor_dir])
    postprocess_components(cfg, inventory, cfg.get_components())

    shutil.copytree(
        cfg.work_dir / "compiled", pkg_path / "compiled", dirs_exist_ok=True
    )

    if not keep_dir:
        shutil.rmtree(temp_dir)
    elif not tmp_dir:
        click.echo(
            f"Compilation working directory: {temp_dir}. Specify `--tmp-dir={temp_dir}`"
            + ", if you want to use local mode for subsequent compilations."
        )


def _setup_inventory(
    inv: Inventory, pkg_name: str, root_class_name: str, value_files: Iterable[Path]
):
    """Setup inventory files which would usually be provided in the cluster's global
    config.
    """
    inv.ensure_dirs()

    tenant_dir = inv.tenant_config_dir("t-fake")
    tenant_dir.mkdir(exist_ok=True)

    # Generate class hierarchy from all additional values files and the target class
    # provided as command line arguments. We use this list as the top-level hierarchy in
    # class `global.commodore`.
    classes = []
    for f in value_files:
        # Add class for value file
        classes.append(f"t-fake.{f.stem}")
        # Symlink value file to inventory
        relsymlink(f, tenant_dir, f.name)

    # Add the package after the additional test config, so that we can provide relevant
    # facts etc. before the package classes are rendered.
    classes.append(f"{pkg_name}.{root_class_name}")

    # setup global.commodore class with class hierarchy containing package target class
    # to compile and any additional value classes provided on the command line.
    global_commodore = inv.global_config_dir / "commodore.yml"
    global_commodore.parent.mkdir(exist_ok=True)
    yaml_dump(
        {
            "classes": classes,
        },
        global_commodore,
    )
    # TODO(sg): Figure out what else we want to provide fallback values for.
    #  The idea is that packages provide test files to override these values if they
    #  need to.
    yaml_dump(
        {
            "parameters": {
                "cluster": {
                    "catalog_url": "ssh://git@git.example.com/org/repo.git",
                    "name": "c-green-test-1234",
                    "tenant": "t-silent-test-1234",
                },
                "facts": {
                    "distribution": "x-fake-distribution",
                    "cloud": "x-fake-cloud",
                    "region": "x-fake-region",
                },
                "argocd": {
                    "namespace": "syn",
                },
                "kapitan": {
                    "secrets": {
                        "vaultkv": {
                            "VAULT_ADDR": "https://vault.syn.example.com",
                            "VAULT_SKIP_VERIFY": "false",
                            "VAULT_CAPATH": "/etc/ssl/certs/",
                            "auth": "token",
                            "engine": "kv-v2",
                            "mount": "clusters/kv",
                        },
                    },
                },
            }
        },
        inv.params_file,
    )

    # Fake Argo CD lib
    # We plug "fake" Argo CD library here because every component relies on it
    # and we don't want to provide it every time when compiling a single component.
    with open(inv.lib_dir / "argocd.libjsonnet", "w", encoding="utf-8") as argocd_libf:
        argocd_libf.write(
            dedent(
                """
            local ArgoApp(component, namespace, project='', secrets=true) = {};
            local ArgoProject(name) = {};

            {
              App: ArgoApp,
              Project: ArgoProject,
            }"""
            )
        )



class K8sObject:
    def __init__(self, obj):
        self._obj = {}
        if obj:
            self._obj = obj
        self._kind = self._obj.get("kind", "")
        self._name = self._obj.get("metadata", {}).get("name", "")
        self._namespace = self._obj.get("metadata", {}).get("namespace", "")

    def __lt__(self, other):
        if self._kind != other._kind:
            return self._kind < other._kind
        if self._namespace != other._namespace:
            return self._namespace < other._namespace
        return self._name < other._name

    def __gt__(self, other):
        if self._kind != other._kind:
            return self._kind > other._kind
        if self._namespace != other._namespace:
            return self._namespace > other._namespace
        return self._name > other._name

    def __eq__(self, other):
        return (
            self._kind == other._kind
            and self._namespace == other._namespace
            and self._name == other._name
        )

    def __le__(self, other):
        return not self.__gt__(other)

    def __ge__(self, other):
        return not self.__lt__(other)

    def __ne__(self, other):
        return not self.__eq__(other)



from __future__ import annotations

from pathlib import Path
from typing import Any

import click

DEPRECATED_PARAMS = [
    "${customer:name}",
    "${cloud:provider}",
    "${cloud:region}",
    "${cluster:dist}",
]


def _lint_data(file: Path, prefix: str, data: Any) -> int:
    if isinstance(data, dict):
        return _lint_dict(file, prefix, data)
    if isinstance(data, list):
        return _lint_list(file, prefix, data)
    if isinstance(data, str):
        return _lint_string(file, prefix, data)

    # Only dicts, lists, or strings can contain parameter references
    return 0


def _lint_string(file: Path, prefix: str, data: str) -> int:
    errcount = 0

    for p in DEPRECATED_PARAMS:
        if p in data:
            click.secho(
                f"> Field '{prefix}' in file '{file}' contains deprecated parameter '{p}'",
                fg="red",
            )
            errcount += 1

    return errcount


def _lint_list(file: Path, prefix: str, data: list) -> int:
    errcount = 0
    for i, v in enumerate(data):
        errcount += _lint_data(file, prefix + f"[{i}]", v)

    return errcount


def _lint_dict(file: Path, prefix: str, data: dict[str, Any]) -> int:
    errcount = 0
    if prefix != "":
        prefix = f"{prefix}."
    for k, v in data.items():
        errcount += _lint_data(file, prefix + k, v)

    return errcount


def lint_deprecated_parameters(file: Path, filecontents: dict[str, Any]) -> int:
    prefix = ""
    return _lint_dict(file, prefix, filecontents)



from pathlib import Path
from typing import Optional

from commodore.multi_dependency import MultiDependency
from commodore.gitrepo import GitRepo


class Package:
    """
    Class representing a config package.
    Used to abstract from details of cloning/checking out the correct package repo and
    version
    """

    _gitrepo: Optional[GitRepo]

    @classmethod
    def clone(cls, cfg, clone_url: str, name: str, version: str = "master"):
        pdep = MultiDependency(clone_url, cfg.inventory.dependencies_dir)
        p = Package(
            name,
            pdep,
            package_dependency_dir(cfg.work_dir, name),
            version=version,
        )
        p.checkout()
        return p

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        name: str,
        dependency: MultiDependency,
        target_dir: Path,
        version: Optional[str] = None,
        sub_path: str = "",
    ):
        self._name = name
        self._version = version
        self._sub_path = sub_path
        self._dependency = dependency
        self._dependency.register_package(name, target_dir)
        self._gitrepo = None

    @property
    def url(self) -> Optional[str]:
        return self._dependency.url

    @property
    def version(self) -> Optional[str]:
        return self._version

    @property
    def sub_path(self) -> str:
        return self._sub_path

    @property
    def repository_dir(self) -> Optional[Path]:
        return self._dependency.get_package(self._name)

    @property
    def repo(self) -> Optional[GitRepo]:
        if not self._gitrepo and self.target_dir and self.target_dir.is_dir():
            self._gitrepo = GitRepo(None, self.target_dir)
        return self._gitrepo

    @property
    def target_dir(self) -> Optional[Path]:
        worktree = self._dependency.get_package(self._name)
        if not worktree:
            return None

        return worktree / self._sub_path

    def checkout(self):
        self._dependency.checkout_package(self._name, self._version)


def package_dependency_dir(work_dir: Path, pname: str) -> Path:
    return work_dir / "dependencies" / f"pkg.{pname}"



import shutil

create_lib = "{{ cookiecutter.add_lib }}" == "y"
add_golden = "{{ cookiecutter.add_golden }}" == "y"

if not create_lib:
    shutil.rmtree("lib")

if not add_golden:
    shutil.rmtree("tests/golden")



from __future__ import annotations

from pathlib import Path
from typing import Optional

from url_normalize.tools import deconstruct_url

from commodore.gitrepo import GitRepo, normalize_git_url


class MultiDependency:
    _repo: GitRepo
    _components: dict[str, Path]
    _packages: dict[str, Path]

    def __init__(self, repo_url: str, dependencies_dir: Path):
        repo_dir = dependency_dir(dependencies_dir, repo_url)
        self._repo = GitRepo(repo_url, repo_dir, bare=True)
        self._components = {}
        self._packages = {}

    @property
    def url(self) -> str:
        return self._repo.remote

    @url.setter
    def url(self, repo_url: str):
        self._repo.remote = repo_url

    @property
    def repo_directory(self) -> Path:
        return Path(self._repo.repo.common_dir).resolve().absolute()

    def get_component(self, name: str) -> Optional[Path]:
        return self._components.get(name)

    def register_component(self, name: str, target_dir: Path):
        if name in self._components:
            raise ValueError(f"component {name} already registered")

        self._components[name] = target_dir

    def deregister_component(self, name: str):
        try:
            del self._components[name]
        except KeyError as e:
            raise ValueError(f"can't deregister unknown component {name}") from e

    def checkout_component(self, name: str, version: str):
        """Create or update worktree for component `name`."""
        target_dir = self.get_component(name)
        if not target_dir:
            raise ValueError(f"can't checkout unknown component {name}")
        self._repo.checkout_worktree(target_dir, version=version)

    def get_package(self, name: str) -> Optional[Path]:
        return self._packages.get(name)

    def register_package(self, name: str, target_dir: Path):
        if name in self._packages:
            raise ValueError(f"package {name} already registered")

        self._packages[name] = target_dir

    def deregister_package(self, name: str):
        try:
            del self._packages[name]
        except KeyError as e:
            raise ValueError(f"can't deregister unknown package {name}") from e

    def checkout_package(self, name: str, version: str):
        """Create or update worktree for package `name`."""
        target_dir = self.get_package(name)
        if not target_dir:
            raise ValueError(f"can't checkout unknown package {name}")
        self._repo.checkout_worktree(target_dir, version=version)

    def initialize_worktree(self, target_dir: Path) -> None:
        """Initialize a worktree in `target_dir`."""
        self._repo.initialize_worktree(target_dir)

    def has_checkouts(self) -> bool:
        return len(self._repo.worktrees) > 1


def dependency_dir(base_dir: Path, repo_url: str) -> Path:
    return base_dir / ".repos" / dependency_key(repo_url)


def dependency_key(repo_url: str) -> str:
    """Create normalized and scheme-agnostic key for the given repo URL.

    This is also used to determine the subpath where the bare checkout is created."""
    repo_url = normalize_git_url(repo_url)
    url_parts = deconstruct_url(repo_url)
    depkey = ""
    if url_parts.host:
        depkey = f"{url_parts.host}/"
    return depkey + url_parts.path[1:]



from __future__ import annotations

import os
import json
from collections.abc import Iterable
from pathlib import Path
from subprocess import call  # nosec

import click

from commodore.config import Config
from commodore.helpers import relsymlink


def jsonnet_dependencies(config: Config) -> Iterable:
    """
    Creates a list of Jsonnet dependencies for the given Components.
    """
    dependencies = []

    for component in config.get_components().values():
        dependencies.append(
            {
                "source": {
                    "local": {
                        "directory": os.path.relpath(
                            # TODO: Update again when we have proper monorepo handling
                            component.repo_directory,
                            start=config.work_dir,
                        ),
                    }
                }
            }
        )

    # Defining the `lib` folder as a local dependency is just a cheap way to have a symlink to that
    # folder.
    dependencies.append(
        {
            "source": {
                "local": {
                    "directory": os.path.relpath(
                        config.inventory.lib_dir, start=config.work_dir
                    ),
                }
            }
        }
    )

    return dependencies


def write_jsonnetfile(file: Path, deps: Iterable):
    """
    Writes the file `jsonnetfile.json` containing all provided dependencies.
    """
    data = {
        "version": 1,
        "dependencies": deps,
        "legacyImports": True,
    }

    with open(file, "w", encoding="utf-8") as f:
        f.write(json.dumps(data, indent=4))
        f.write("\n")


def fetch_jsonnet_libraries(cwd: Path, deps: Iterable = None):
    """
    Download Jsonnet libraries using Jsonnet-Bundler.
    """
    jsonnetfile = cwd / "jsonnetfile.json"
    if not jsonnetfile.exists() or deps:
        if not deps:
            deps = []
        write_jsonnetfile(jsonnetfile, deps)

    inject_essential_libraries(jsonnetfile)

    try:
        # To make sure we don't use any stale lock files
        lock_file = cwd / "jsonnetfile.lock.json"
        if lock_file.exists():
            lock_file.unlink()
        if call(["jb", "install"], cwd=cwd) != 0:
            raise click.ClickException("jsonnet-bundler exited with error")
    except FileNotFoundError as e:
        raise click.ClickException(
            "the jsonnet-bundler executable `jb` could not be found"
        ) from e

    # Link essential libraries for backwards compatibility.
    lib_dir = (cwd / "vendor" / "lib").resolve()
    lib_dir.mkdir(exist_ok=True)
    relsymlink(
        cwd / "vendor" / "kube-libsonnet" / "kube.libsonnet", lib_dir, "kube.libjsonnet"
    )


def inject_essential_libraries(file: Path):
    """
    Ensures essential libraries are added to `jsonnetfile.json`.
    :param file: The path to `jsonnetfile.json`.
    """
    with open(file, "r", encoding="utf-8") as f:
        data = json.load(f)

    deps = data["dependencies"]
    has_kube = False
    for dep in deps:
        remote = dep.get("source", {}).get("git", {}).get("remote", "")
        has_kube = has_kube or "kube-libsonnet" in remote

    if not has_kube:
        deps.append(
            {
                "source": {
                    "git": {"remote": "https://github.com/bitnami-labs/kube-libsonnet"}
                },
                "version": "v1.19.0",
            },
        )

    with open(file, "w", encoding="utf-8") as j:
        json.dump(data, j, indent=4)
        j.write("\n")



from os import makedirs
from pathlib import Path as P
from typing import Union

from commodore.component import Component


class Inventory:
    _work_dir: P

    def __init__(self, work_dir: P = None):
        if work_dir:
            self._work_dir = work_dir
        else:
            self._work_dir = P(".")

    @property
    def work_dir(self) -> P:
        return self._work_dir

    @work_dir.setter
    def work_dir(self, d: P):
        self._work_dir = d

    @property
    def inventory_dir(self) -> P:
        return self._work_dir / "inventory"

    @property
    def dependencies_dir(self) -> P:
        return self._work_dir / "dependencies"

    @property
    def classes_dir(self) -> P:
        return self.inventory_dir / "classes"

    @property
    def components_dir(self) -> P:
        return self.classes_dir / "components"

    @property
    def defaults_dir(self) -> P:
        return self.classes_dir / "defaults"

    @property
    def targets_dir(self) -> P:
        return self.inventory_dir / "targets"

    @property
    def lib_dir(self) -> P:
        return self.dependencies_dir / "lib"

    @property
    def libs_dir(self) -> P:
        return self.dependencies_dir / "libs"

    @property
    def global_config_dir(self) -> P:
        return self.classes_dir / "global"

    @property
    def bootstrap_target(self) -> str:
        return "cluster"

    @property
    def params_dir(self) -> P:
        return self.classes_dir / "params"

    @property
    def params_file(self) -> P:
        return self.params_dir / f"{self.bootstrap_target}.yml"

    @property
    def output_dir(self) -> P:
        return self._work_dir / "compiled"

    def tenant_config_dir(self, tenant: str) -> P:
        return self.classes_dir / tenant

    def package_dir(self, pkgname: str) -> P:
        return self.classes_dir / pkgname

    # pylint: disable=unsubscriptable-object
    def component_file(self, component: Union[Component, str]) -> P:
        return self.components_dir / f"{_component_name(component)}.yml"

    # pylint: disable=unsubscriptable-object
    def defaults_file(self, component: Union[Component, str]) -> P:
        return self.defaults_dir / f"{_component_name(component)}.yml"

    # pylint: disable=unsubscriptable-object
    def target_file(self, target: Union[Component, str]) -> P:
        return self.targets_dir / f"{_component_name(target)}.yml"

    def ensure_dirs(self):
        makedirs(self.components_dir, exist_ok=True)
        makedirs(self.defaults_dir, exist_ok=True)
        makedirs(self.params_dir, exist_ok=True)
        makedirs(self.lib_dir, exist_ok=True)
        makedirs(self.libs_dir, exist_ok=True)
        makedirs(self.targets_dir, exist_ok=True)


# pylint: disable=unsubscriptable-object
def _component_name(component: Union[Component, str]) -> str:
    if isinstance(component, Component):
        return component.name

    return component



"""
Commodore. Build dynamic inventories and compile catalogs with Kapitan
"""
from pathlib import Path as P
from importlib_metadata import version

__url__ = "https://github.com/projectsyn/commodore/"
__git_version__ = 'v1.7.0'
__version__ = version("syn-commodore")

# provide Commodore installation dir as variable that can be imported
__install_dir__ = P(__file__).parent



from __future__ import annotations

import random
import time

from pathlib import Path
from typing import Iterable, Optional

import click
import git
import github
import yaml.parser

from github.Repository import Repository

from commodore.config import Config
from commodore.helpers import yaml_load

from . import Package
from .template import PackageTemplater


def sync_packages(
    config: Config,
    package_list: Path,
    dry_run: bool,
    pr_branch: str,
    pr_label: Iterable[str],
) -> None:
    if not config.github_token:
        raise click.ClickException("Can't continue, missing GitHub API token.")

    try:
        pkgs = yaml_load(package_list)
        if not isinstance(pkgs, list):
            typ = type(pkgs)
            raise ValueError(f"unexpected type: {typ}")
    except ValueError as e:
        raise click.ClickException(f"Expected a list in '{package_list}', but got {e}")
    except (yaml.parser.ParserError, yaml.scanner.ScannerError):
        raise click.ClickException(f"Failed to parse YAML in '{package_list}'")

    gh = github.Github(config.github_token)
    pkg_count = len(pkgs)
    for i, pn in enumerate(pkgs, start=1):
        click.secho(f"Synchronizing {pn}", bold=True)
        porg, preponame = pn.split("/")
        pname = preponame.replace("package-", "", 1)

        # Clone package
        try:
            gr = gh.get_repo(pn)
        except github.UnknownObjectException:
            click.secho(f" > Repository {pn} doesn't exist, skipping...", fg="yellow")
            continue
        p = Package.clone(config, gr.clone_url, pname, version=gr.default_branch)

        if not (p.target_dir / ".cruft.json").is_file():
            click.echo(f" > Skipping repo {pn} which doesn't have `.cruft.json`")
            continue

        # Run `package update`
        t = PackageTemplater.from_existing(config, p.target_dir)
        changed = t.update(print_completion_message=False)

        # Create or update PR if there were updates
        if changed:
            ensure_branch(p, pr_branch)
            cu = ensure_pr(p, pn, gr, dry_run, pr_branch, pr_label)
            click.secho(f"PR for package {pn} successfully {cu}", bold=True)

            if i < pkg_count:
                # except when processing the last package in the list, sleep for 1-2
                # seconds to avoid hitting secondary rate-limits for PR creation. No
                # need to sleep if we're not creating a PR.
                # Without the #nosec annotations bandit warns (correctly) that
                # `random.random()` generates weak random numbers, but since the quality
                # of the randomness doesn't matter here, we don't need to use a more
                # expensive RNG.
                backoff = 1.0 + random.random()  # nosec
                time.sleep(backoff)
        else:
            click.secho(f"Package {pn} already up-to-date", bold=True)


def message_body(c: git.objects.commit.Commit) -> str:
    if isinstance(c.message, bytes):
        msg = str(c.message, encoding="utf-8")
    else:
        msg = c.message
    paragraphs = msg.split("\n\n")
    return "\n\n".join(paragraphs[1:])


def ensure_branch(p: Package, branch_name: str):
    """Create or reset `template-sync` branch pointing to our new template update
    commit."""
    if not p.repo:
        raise ValueError("package repo not initialized")
    r = p.repo.repo
    has_sync_branch = any(h.name == branch_name for h in r.heads)

    if not has_sync_branch:
        r.create_head(branch_name)
    else:
        new_update = r.head.commit
        template_sync = [h for h in r.heads if h.name == branch_name][0]
        template_sync.set_reference(new_update)


def ensure_pr(
    p: Package,
    pn: str,
    gr: Repository,
    dry_run: bool,
    branch_name: str,
    pr_labels: Iterable[str],
) -> Optional[str]:
    """Create or update template sync PR."""
    if not p.repo:
        raise ValueError("package repo not initialized")

    prs = gr.get_pulls(state="open")
    has_sync_pr = any(pr.head.ref == branch_name for pr in prs)

    cu = "update" if has_sync_pr else "create"
    if dry_run:
        click.echo(f"Would {cu} PR for {pn}")
        return None

    r = p.repo.repo
    r.remote().push(branch_name, force=True)
    pr_body = message_body(r.head.commit)

    try:
        if not has_sync_pr:
            sync_pr = gr.create_pull(
                "Update from package template",
                pr_body,
                gr.default_branch,
                branch_name,
            )
        else:
            sync_pr = [pr for pr in prs if pr.head.ref == branch_name][0]
            sync_pr.edit(body=pr_body)
        sync_pr.add_to_labels(*list(pr_labels))
    except github.UnknownObjectException:
        click.echo(
            f"Unable to {cu} PR for {pn}. "
            + "Please make sure your GitHub token has permission 'public_repo'"
        )
        return None

    return f"{cu}d"



from __future__ import annotations

import json

from pathlib import Path
from typing import Any, Optional, Sequence

import click


from commodore.config import Config
from commodore.cruft._commands import create as cruft_create, update as cruft_update
from commodore.dependency_mgmt.discovery import (
    RESERVED_PACKAGE_PATTERN,
    TENANT_PREFIX_PATTERN,
)
from commodore.dependency_templater import Templater, Renderer
from commodore.package import package_dependency_dir


# pylint: disable=too-many-instance-attributes
class PackageTemplater(Templater):
    template_url: str
    template_version: Optional[str] = None
    _test_cases: list[str] = ["defaults"]
    copyright_year: Optional[str] = None
    _target_dir: Optional[Path] = None

    @classmethod
    def from_existing(cls, config: Config, package_path: Path):
        if not package_path.is_dir():
            raise click.ClickException("Provided package path isn't a directory")
        with open(package_path / ".cruft.json", encoding="utf-8") as cfg:
            cruft_json = json.load(cfg)

        cookiecutter_args = cruft_json["context"]["cookiecutter"]
        t = PackageTemplater(
            config, cookiecutter_args["slug"], name=cookiecutter_args["name"]
        )
        t._target_dir = package_path
        t.output_dir = package_path.absolute().parent
        t.template_url = cruft_json["template"]
        t.template_version = cruft_json.get("checkout")

        if "test_cases" in cookiecutter_args:
            t.test_cases = cookiecutter_args["test_cases"].split(" ")
        t.golden_tests = cookiecutter_args["add_golden"] == "y"
        t.github_owner = cookiecutter_args["github_owner"]
        t.copyright_holder = cookiecutter_args["copyright_holder"]
        t.copyright_year = cookiecutter_args["copyright_year"]
        return t

    @property
    def template_commit(self) -> str:
        with open(self.target_dir / ".cruft.json", "r", encoding="utf-8") as f:
            cruft_json = json.load(f)
            return cruft_json["commit"]

    @property
    def test_cases(self) -> list[str]:
        """Return list of test cases.

        The getter deduplicates the stored list before returning it.

        Don't use `append()` on the returned list to add test cases to the package, as
        the getter returns a copy of the list stored in the object."""
        cases = []
        for t in self._test_cases:
            if t not in cases:
                cases.append(t)
        return cases

    @test_cases.setter
    def test_cases(self, test_cases: list[str]):
        self._test_cases = test_cases

    def _cruft_renderer(
        self,
        template_location: str,
        extra_context: dict[str, Any],
        no_input: bool,
        output_dir: Path,
    ):
        cruft_create(
            template_location,
            checkout=self.template_version,
            extra_context=extra_context,
            no_input=no_input,
            output_dir=output_dir,
        )

    def _validate_slug(self, value: str):
        # First perform default slug checks
        slug = super()._validate_slug(value)
        # Then perform additional checks for package slug
        if RESERVED_PACKAGE_PATTERN.match(slug):
            raise click.ClickException(f"Package can't use reserved slug '{slug}'")
        if TENANT_PREFIX_PATTERN.match(slug):
            raise click.ClickException(
                "Package slug can't use reserved tenant prefix 't-'"
            )
        return slug

    @property
    def cookiecutter_args(self) -> dict[str, str]:
        return {
            "add_golden": "y" if self.golden_tests else "n",
            "copyright_holder": self.copyright_holder,
            "copyright_year": (
                self.today.strftime("%Y")
                if not self.copyright_year
                else self.copyright_year
            ),
            "github_owner": self.github_owner,
            "name": self.name,
            "slug": self.slug,
            # The template expects the test cases in a single string separated by
            # spaces.
            "test_cases": " ".join(self.test_cases),
        }

    @property
    def deptype(self) -> str:
        return "package"

    @property
    def template_renderer(self) -> Renderer:
        return self._cruft_renderer

    @property
    def target_dir(self) -> Path:
        if self._target_dir:
            return self._target_dir

        if self.output_dir:
            return self.output_dir / self.slug

        return package_dependency_dir(self.config.work_dir, self.slug)

    @property
    def template(self) -> str:
        return self.template_url

    @property
    def additional_files(self) -> Sequence[str]:
        return [
            ".github",
            ".gitignore",
            ".*.yml",
            ".editorconfig",
            ".cruft.json",
        ]

    def update(self, print_completion_message: bool = True) -> bool:
        cruft_update(
            self.target_dir,
            cookiecutter_input=False,
            checkout=self.template_version,
            extra_context=self.cookiecutter_args,
        )

        updated = self.commit(
            "Update from template\n\n"
            + f"Template version: {self.template_version} ({self.template_commit[:7]})",
            init=False,
        )

        if print_completion_message:
            if updated:
                click.secho(
                    f"{self.deptype.capitalize()} {self.name} successfully updated 🎉",
                    bold=True,
                )
            else:
                click.secho(
                    f"{self.deptype.capitalize()} {self.name} already up-to-date 🎉",
                    bold=True,
                )

        return updated



from __future__ import annotations

import os
from collections.abc import Iterable
from enum import Enum
from pathlib import Path
from typing import Any, Optional

from kapitan.reclass import reclass
from kapitan.reclass.reclass import settings as reclass_settings
from kapitan.reclass.reclass import core as reclass_core
from kapitan.reclass.reclass import errors as reclass_errors

from commodore.cluster import Cluster, render_params
from commodore.component import component_parameters_key
from commodore.helpers import yaml_dump, yaml_load
from commodore.inventory import Inventory

FAKE_TENANT_ID = "t-foo"
FAKE_CLUSTER_ID = "c-bar"


class ClassNotFound(reclass_errors.ClassNotFound):
    """
    Create a wrapper around kapitan.reclass's ClassNotFound exception.

    This allows us to have clean exception handling outside of this file, because we
    unfortunately can't catch ClassNotFound directly in our code if we import
    and use Kapitan reclass's ClassNotFound with any import which doesn't map the class
    to `reclass.errors.ClassNotFound`. This is the case because Python's implementation
    doesn't try to figure out if two imports are the same if they have different import
    paths, cf. https://docs.python.org/3/reference/import.html.

    The only import that should work (in my understanding) would be
    `from kapitan.reclass import reclass` but with that import we get the error
    "AttributeError: module 'kapitan.reclass.reclass' has no attribute 'errors'"
    """

    @classmethod
    def from_reclass(cls, e: reclass_errors.ClassNotFound):
        """Wrap a reclass.errors.ClassNotFound instance in our wrapper."""
        return ClassNotFound(e.storage, e.name, e.path)


class InventoryFacts:
    def __init__(
        self,
        global_config: str,
        tenant_config: Optional[str],
        extra_class_files: Iterable[Path],
        allow_missing_classes: bool,
        ignore_class_notfound_warning: bool = True,
    ):
        self._global_config = global_config
        self._tenant_config = tenant_config
        self._extra_class_files = extra_class_files
        self._allow_missing_classes = allow_missing_classes
        self._ignore_class_notfound_warning = ignore_class_notfound_warning

    @property
    def global_config(self) -> str:
        return self._global_config

    @property
    def tenant_config(self) -> Optional[str]:
        return self._tenant_config

    @property
    def extra_classes(self) -> list[str]:
        return [cf.stem for cf in self._extra_class_files]

    @property
    def extra_class_files(self) -> Iterable[Path]:
        return self._extra_class_files

    @property
    def allow_missing_classes(self) -> bool:
        return self._allow_missing_classes

    @property
    def ignore_class_notfound_warning(self) -> bool:
        return self._ignore_class_notfound_warning

    @property
    def tenant_id(self) -> str:
        tenant_id = None
        for f in self.extra_class_files:
            fc = yaml_load(f)
            tenant_id = (
                fc.get("parameters", {}).get("cluster", {}).get("tenant", tenant_id)
            )

        if not tenant_id:
            tenant_id = FAKE_TENANT_ID

        return tenant_id

    @property
    def cluster_id(self) -> str:
        cluster_id = None
        for f in self.extra_class_files:
            fc = yaml_load(f)
            cluster_id = (
                fc.get("parameters", {}).get("cluster", {}).get("name", cluster_id)
            )

        if not cluster_id:
            cluster_id = FAKE_CLUSTER_ID

        return cluster_id


class InventoryParameters:
    def __init__(self, inv):
        self._inventory = inv

    @property
    def distribution(self):
        return self._inventory["parameters"]["facts"]["distribution"]

    @property
    def cloud(self):
        return self._inventory["parameters"]["facts"]["cloud"]

    @property
    def region(self):
        return self._inventory["parameters"]["facts"]["region"]

    def parameters(self, param: Optional[str] = None):
        params = self._inventory["parameters"]
        if param is not None:
            params = params.get(component_parameters_key(param), {})

        return params

    @property
    def applications(self):
        return self._inventory["applications"]


class DefaultsFact(Enum):
    DISTRIBUTION = "distribution"
    CLOUD = "cloud"
    REGION = "region"


class InventoryFactory:
    def __init__(self, work_dir: Path, global_dir: Path, tenant_dir: Optional[Path]):
        self._global_dir = global_dir
        self._tenant_dir = tenant_dir
        self._inventory = Inventory(work_dir=work_dir)
        self._distributions = self._find_values(DefaultsFact.DISTRIBUTION)
        self._clouds = self._find_values(DefaultsFact.CLOUD)
        self._cloud_regions: dict[str, Iterable[str]] = {}
        for cloud in self._clouds:
            r = self._find_values(DefaultsFact.REGION, cloud=cloud)
            self._cloud_regions[cloud] = [it for it in r if it != "params"]

    @property
    def directory(self) -> Path:
        return self._inventory.inventory_dir

    @property
    def classes_dir(self) -> Path:
        return self._inventory.classes_dir

    @property
    def targets_dir(self) -> Path:
        return self._inventory.targets_dir

    @property
    def global_dir(self) -> Path:
        return self._global_dir

    @property
    def tenant_dir(self) -> Optional[Path]:
        return self._tenant_dir

    def _reclass_config(
        self, allow_missing_classes: bool, ignore_class_notfound_warning: bool = True
    ) -> dict:
        return {
            "storage_type": "yaml_fs",
            "inventory_base_uri": str(self.directory.absolute()),
            "nodes_uri": str(self.targets_dir.absolute()),
            "classes_uri": str(self.classes_dir.absolute()),
            "compose_node_name": False,
            "allow_none_override": True,
            "ignore_class_notfound": allow_missing_classes,
            "ignore_class_notfound_warning": ignore_class_notfound_warning,
        }

    def _render_inventory(
        self,
        target: Optional[str] = None,
        allow_missing_classes: bool = True,
        ignore_class_notfound_warning: bool = True,
    ) -> dict[str, Any]:
        rc = self._reclass_config(
            allow_missing_classes,
            ignore_class_notfound_warning=ignore_class_notfound_warning,
        )
        storage = reclass.get_storage(
            rc["storage_type"],
            rc["nodes_uri"],
            rc["classes_uri"],
            rc["compose_node_name"],
        )
        class_mappings = rc.get("class_mappings")
        _reclass = reclass_core.Core(
            storage, class_mappings, reclass_settings.Settings(rc)
        )

        try:
            if target:
                return _reclass.nodeinfo(target)

            return _reclass.inventory()
        except Exception as e:
            if type(e).__name__ == "ClassNotFound":
                # Wrap Kapitan reclass's ClassNotFound with ours so we can cleanly
                # catch it. See docsstring for our ClassNotFound for a more detailed
                # explanation why this is necessary.
                raise ClassNotFound.from_reclass(e)
            raise

    def reclass(self, invfacts: InventoryFacts) -> InventoryParameters:
        cluster_response: dict[str, Any] = {
            "id": invfacts.cluster_id,
            "tenant": invfacts.tenant_id,
            "displayName": "Foo Inc. Bar cluster",
            "facts": {
                "distribution": "x-fake-distribution",
                "cloud": "x-fake-cloud",
                "region": "x-fake-region",
                "lieutenant-instance": "lieutenant-prod",
            },
            "gitRepo": {
                "url": "not-a-real-repo",
            },
        }
        cluster = Cluster(
            cluster_response=cluster_response,
            tenant_response={
                "id": invfacts.tenant_id,
                "displayName": "Foo Inc.",
                "gitRepo": {
                    "url": "not-a-real-repo",
                },
            },
        )
        params = render_params(self._inventory, cluster)

        yaml_dump(params, self.classes_dir / "base.yml")

        # Create the following fake hierarchy for the render target:
        # classes:
        # - base
        # - user-supplied-file-1
        # - user-supplied-file-2
        # - ...
        # - global.commodore
        classes = ["base"] + invfacts.extra_classes + ["global.commodore"]
        yaml_dump(
            {
                "classes": classes,
                "parameters": {
                    "_instance": "global",
                },
            },
            self.targets_dir / "global.yml",
        )

        if not invfacts.tenant_config:
            # Create fake cluster class, this allows rendering the inventory with
            # allow-missing-classes=False to work when no tenant is specified.
            yaml_dump(
                {},
                self.classes_dir / FAKE_TENANT_ID / f"{FAKE_CLUSTER_ID}.yml",
            )

        return InventoryParameters(
            self._render_inventory(
                "global",
                allow_missing_classes=invfacts.allow_missing_classes,
                ignore_class_notfound_warning=invfacts.ignore_class_notfound_warning,
            )
        )

    def _find_values(self, fact: DefaultsFact, cloud: str = None) -> Iterable[str]:
        values = []
        value_path = self.global_dir / fact.value
        if fact == DefaultsFact.REGION:
            if not cloud:
                raise ValueError(f"cloud must not be None if fact is {fact}")
            value_path = self.global_dir / "cloud" / cloud
        if value_path.is_dir():
            for f in value_path.iterdir():
                if f.is_file() and f.suffix in (".yml", ".yaml"):
                    values.append(f.stem)
        return values

    @property
    def distributions(self) -> Iterable[str]:
        return self._distributions

    @property
    def clouds(self) -> Iterable[str]:
        return self._clouds

    @property
    def cloud_regions(self) -> dict[str, Iterable[str]]:
        return self._cloud_regions

    @classmethod
    def from_repo_dirs(
        cls,
        work_dir: Path,
        global_dir: Path,
        tenant_dir: Optional[Path],
        invfacts: InventoryFacts,
    ):
        classes_dir = work_dir / "inventory" / "classes"
        os.makedirs(work_dir / "inventory" / "targets", exist_ok=True)
        os.makedirs(classes_dir, exist_ok=True)
        if not tenant_dir:
            os.makedirs(
                work_dir / "inventory" / "classes" / FAKE_TENANT_ID, exist_ok=True
            )
        os.symlink(global_dir.absolute(), classes_dir / "global")
        if tenant_dir:
            os.symlink(tenant_dir.absolute(), classes_dir / invfacts.tenant_id)
        for cf in invfacts.extra_class_files:
            os.symlink(cf.absolute(), classes_dir / cf.name)
        return InventoryFactory(
            work_dir=work_dir,
            global_dir=classes_dir / "global",
            tenant_dir=classes_dir / invfacts.tenant_id,
        )



from __future__ import annotations

from pathlib import Path
from typing import Any

import click
from commodore.dependency_mgmt.version_parsing import DepType


def lint_dependency_specification(
    deptype: DepType, file: Path, filecontents: dict[str, Any]
) -> int:
    errcount = 0
    components = filecontents.get("parameters", {}).get(deptype.value, {})
    deptype_str = deptype.name.capitalize()
    for d, dspec in components.items():
        if "version" not in dspec:
            click.secho(
                f"> {deptype_str} specification for {d} "
                + f"is missing key 'version' in {file}",
                fg="red",
            )
            errcount += 1

        unk_keys = set(dspec.keys()) - {"url", "version"}
        if len(unk_keys) > 0:
            click.secho(
                f"> {deptype_str} specification for {d} "
                + f"contains unknown key(s) '{unk_keys}' in {file}",
                fg="red",
            )
            errcount += 1

        durl = dspec.get("url", "")
        if not isinstance(durl, str):
            click.secho(
                f"> {deptype_str} {d} url is of type {type(durl).__name__}"
                + f" (expected string) in {file}",
                fg="red",
            )
            errcount += 1

        dversion = dspec.get("version", "")
        if not isinstance(dversion, str):
            click.secho(
                f"> {deptype_str} {d} version is of type {type(dversion).__name__},"
                + f" (expected string) in {file}",
                fg="red",
            )
            errcount += 1

    return errcount


def lint_components(file: Path, filecontents: dict[str, Any]) -> int:
    return lint_dependency_specification(DepType.COMPONENT, file, filecontents)


def lint_packages(file: Path, filecontents: dict[str, Any]) -> int:
    return lint_dependency_specification(DepType.PACKAGE, file, filecontents)



from __future__ import annotations

from pathlib import Path as P
from typing import Any, Callable, ClassVar, Protocol

import click

from commodore.config import Config, Component

from .jsonnet import run_jsonnet_filter, validate_jsonnet_filter
from .builtin_filters import run_builtin_filter, validate_builtin_filter


class FilterFunc(Protocol):
    # pylint: disable=too-many-arguments
    def __call__(
        self,
        config: Config,
        inv: dict,
        component: Component,
        instance: str,
        filterid: str,
        path: P,
        **filterargs: str,
    ):
        ...


ValidateFunc = Callable[[Config, Component, str, dict], dict]


class Filter:
    type: str
    filter: str
    path: P
    filterargs: dict
    enabled: bool

    # PyLint complains about ClassVar not being subscriptable
    # pylint: disable=unsubscriptable-object
    _run_handlers: ClassVar[dict[str, FilterFunc]] = {
        "builtin": run_builtin_filter,
        "jsonnet": run_jsonnet_filter,
    }
    # pylint: disable=unsubscriptable-object
    _validate_handlers: ClassVar[dict[str, ValidateFunc]] = {
        "builtin": validate_builtin_filter,
        "jsonnet": validate_jsonnet_filter,
    }
    # pylint: disable=unsubscriptable-object
    _required_keys: ClassVar[set[str]] = {"type", "path", "filter"}

    def __init__(self, fd: dict):
        """
        Assumes that `fd` has been validated with `_validate_filter`.
        """
        self.type = fd["type"]
        self.filter = fd["filter"]
        self.path = P(fd["path"])
        self.enabled = fd.get("enabled", True)
        self.filterargs = fd.get("filterargs", {})
        self._runner: FilterFunc = self._run_handlers[self.type]

    def run(self, config: Config, inventory: dict, component: Component, instance: str):
        """
        Run the filter.
        """
        if not self.enabled:
            click.secho(
                f" > Skipping disabled filter {self.filter} on path {self.path}"
            )
            return

        self._runner(
            config,
            inventory,
            component,
            instance,
            self.filter,
            self.path,
            **self.filterargs,
        )

    @classmethod
    def validate(cls, config: Config, c: Component, instance: str, f: dict):
        """
        Validate filter definition in `f`.
        Raises exceptions as appropriate when the definition is invalid.
        Returns the definiton if it validates successfully.
        """
        if not all(key in f for key in cls._required_keys):
            missing_required_keys = cls._required_keys - f.keys()
            raise KeyError(f"Filter is missing required key(s) {missing_required_keys}")

        if "enabled" in f and not isinstance(f["enabled"], bool):
            raise ValueError("Filter key 'enabled' is not a boolean")

        typ = f["type"]
        if typ not in cls._validate_handlers:
            raise ValueError(f"Filter has unknown type {typ}")

        # perform type-specific extra validation
        cls._validate_handlers[typ](config, c, instance, f)

        return f

    @classmethod
    def from_dict(cls, config: Config, c: Component, instance: str, f: dict):
        """
        Create Filter object from filter definition dict `f`.
        Raises exceptions as appropriate when the definition is invalid.
        Returns a Filter object if the passed definition validates successfully.
        """
        return Filter(Filter.validate(config, c, instance, f))


def _get_inventory_filters(inv: dict[str, Any]) -> list[dict[str, Any]]:
    """
    Return list of filters defined in inventory.

    Inventory filters are expected to be defined as a list in
    `parameters.commodore.postprocess.filters`.
    """
    commodore = inv["parameters"].get("commodore", {})
    return commodore.get("postprocess", {}).get("filters", [])


def postprocess_components(
    config: Config,
    kapitan_inventory: dict[str, dict[str, Any]],
    components: dict[str, Component],
):
    click.secho("Postprocessing...", bold=True)

    aliases = config.get_component_aliases()

    for a, cn in aliases.items():
        c = components[cn]
        inv = kapitan_inventory.get(a)
        if not inv:
            click.echo(f" > No target exists for component {cn}, skipping...")
            continue

        # inventory filters
        invfilters = _get_inventory_filters(inv)

        filters: list[Filter] = []
        for fd in invfilters:
            try:
                filters.append(Filter.from_dict(config, c, a, fd))
            except (KeyError, ValueError) as e:
                filtername = fd.get("filter", "<unknown>")
                click.secho(
                    f" > Skipping filter '{filtername}' with invalid definition {fd}: {e}",
                    fg="yellow",
                )

        if len(filters) > 0 and config.debug:
            click.echo(f" > {cn}...")

        for f in filters:
            if config.debug:
                click.secho(f"   > Executing filter '{f.type}:{f.filter}'")
            f.run(config, inv, c, a)



from __future__ import annotations

import json
import sys
import threading
import time
import webbrowser

from functools import partial
from queue import Queue
from typing import Optional, Any

from http.server import BaseHTTPRequestHandler, HTTPServer
from urllib.parse import urlparse, parse_qs

import click
import jwt
import requests

# pylint: disable=redefined-builtin
from requests.exceptions import ConnectionError, HTTPError

from oauthlib.oauth2 import WebApplicationClient

from .config import Config
from . import tokencache


class OIDCError(Exception):
    pass


class OIDCCallbackServer:
    done_queue: Queue = Queue()
    thread: threading.Thread
    server: HTTPServer

    def __init__(
        self,
        client: WebApplicationClient,
        token_url: str,
        lieutenant_url: Optional[str],
    ):
        self.client = client
        self.token_endpoint = token_url
        self.lieutenant_url = lieutenant_url

        handler = partial(
            OIDCCallbackHandler, client, token_url, lieutenant_url, self.done_queue
        )

        self.server = HTTPServer(("", 18000), handler)
        self.thread = threading.Thread(target=self.server.serve_forever)
        self.thread.daemon = True

    def start(self):
        self.thread.start()

    def join(self):
        self.done_queue.get()
        self.server.shutdown()
        self.thread.join()


class OIDCCallbackHandler(BaseHTTPRequestHandler):
    client: WebApplicationClient
    done: Queue

    token_url: str
    redirect_url: str

    lieutenant_url: Optional[str]

    def __init__(
        self,
        client: WebApplicationClient,
        token_url: str,
        lieutenant_url: Optional[str],
        done_queue: Queue,
        *args,
        **kwargs,
    ):
        self.client = client
        self.done = done_queue
        self.lieutenant_url = lieutenant_url
        self.token_url = token_url
        self.redirect_url = "http://localhost:18000"
        super().__init__(*args, **kwargs)

    # pylint: disable=unused-argument
    # pylint: disable=redefined-builtin
    def log_message(self, format, *args):
        return

    def do_GET(self):
        if self.path == "/healthz":
            self.send_response(200)
            self.end_headers()
            self.wfile.write(str.encode("ok"))
            return

        query_components = parse_qs(urlparse(self.path).query)
        if "code" not in query_components or len(query_components["code"]) == 0:
            self.close(422, "invalid callback: no code provided")
            return

        code = query_components["code"][0]
        try:
            tokens = self.get_oidc_tokens(code)
        except (ConnectionError, HTTPError) as e:
            self.close(500, f"failed to connect to IdP: {e}")
            return

        id_token = tokens.get("id_token")

        if id_token is None:
            self.close(500, "no id_token provided")
            return

        if self.lieutenant_url is None:
            print(id_token)
        else:
            try:
                tokencache.save(self.lieutenant_url, tokens)
            except IOError as e:
                self.close(500, f"failed to save token {e}")
                return

        self.close(200, success_page)
        return

    def get_oidc_tokens(self, code) -> dict[str, Any]:
        token_url, headers, body = self.client.prepare_token_request(
            self.token_url,
            redirect_url=self.redirect_url,
            code=code,
        )

        token_response = requests.post(
            token_url,
            headers=headers,
            data=body,
        )
        token_response.raise_for_status()

        return self.client.parse_request_body_response(token_response.text)

    def close(self, code: int, msg: str):
        self.send_response(code)
        self.send_header("Content-type", "text/html")
        self.end_headers()
        self.wfile.write(str.encode(msg))
        self.done.put(True)


success_page = """
<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="UTF-8">
    <title>Authorized</title>
    <script>
        setTimeout(window.close, 5000);
    </script>
     <style>
        body {
            background-color: #eee;
            margin: 0;
            padding: 0;
            font-family: sans-serif;
        }
        .placeholder {
            margin: 2em;
            padding: 2em;
            background-color: #fff;
            border-radius: 1em;
        }
    </style>
</head>
<body>
    <div class="placeholder">
        <h1>Authorized</h1>
        <p>You can close this window.</p>
    </div>
</body>
</html>
"""


def get_idp_cfg(discovery_url: str) -> Any:
    try:
        r = requests.get(discovery_url)
    except ConnectionError as e:
        raise OIDCError(f"Unable to connect to IDP at {discovery_url}") from e
    try:
        resp = json.loads(r.text)
    except json.JSONDecodeError as e:
        raise OIDCError("Client error: Unable to parse JSON") from e
    try:
        r.raise_for_status()
    except HTTPError as e:
        raise OIDCError(f"IDP returned {r.status_code} {e}") from e
    else:
        return resp


def refresh_tokens(
    config: Config, client: WebApplicationClient, token_endpoint: str
) -> bool:
    """Try refreshing the access and id tokens using the current refresh token.

    Tokens are fetched from the token cache based on the data in `config`. The new
    tokens are written to the token cache.

    Returns `True` if refreshing the token was successful."""
    if config.api_url is None:
        # We can't refresh tokens if we don't know the API URL to fetch the old tokens
        # from the cache.
        return False

    tokens = tokencache.get(config.api_url)
    refresh_token = tokens.get("refresh_token")
    if refresh_token is None:
        return False
    # We don't verify the signature, we just want to know if the refresh token is
    # expired.
    try:
        t = jwt.decode(
            refresh_token, algorithms=["RS256"], options={"verify_signature": False}
        )
    except jwt.exceptions.InvalidTokenError:
        # We can't parse the refresh token, notify caller that they need to request a
        # fresh set of tokens.
        return False

    if "exp" in t and t["exp"] > time.time():
        # Only try to refresh the tokens if the refresh token isn't expired yet.
        token_url, headers, body = client.prepare_refresh_token_request(
            token_url=token_endpoint,
            refresh_token=refresh_token,
            client_id=config.oidc_client,
        )
        try:
            token_response = requests.post(token_url, headers=headers, data=body)
            token_response.raise_for_status()
        except (ConnectionError, HTTPError) as e:
            click.echo(f" > Failed to refresh OIDC token with {e}")
            return False

        # If refresh request was successful, parse response and store new
        # tokens in tokencache
        new_tokens = client.parse_request_body_response(token_response.text)
        tokencache.save(config.api_url, new_tokens)
        return True

    return False


def login(config: Config):
    config.discover_oidc_config()

    if config.oidc_client is None:
        raise click.ClickException("Required OIDC client not set")
    if config.oidc_discovery_url is None:
        raise click.ClickException("Required OIDC discovery URL not set")

    if config.api_token:
        # Short-circuit if we already have a valid API token
        return

    client = WebApplicationClient(config.oidc_client)
    idp_cfg = get_idp_cfg(config.oidc_discovery_url)
    if refresh_tokens(config, client, idp_cfg["token_endpoint"]):
        # Short-circuit if refreshing the token was successful.
        return

    # Request new token through login flow if we weren't able to refresh the existing
    # token.
    server = OIDCCallbackServer(client, idp_cfg["token_endpoint"], config.api_url)
    server.start()

    # Wait for server to run
    r = requests.Response()
    r.status_code = 500
    while r.status_code != 200:
        try:
            r = requests.get("http://localhost:18000/healthz")
        except ConnectionError:
            pass

    request_uri = client.prepare_request_uri(
        idp_cfg["authorization_endpoint"],
        redirect_uri="http://localhost:18000",
        scope=["openid", "email", "profile"],
    )
    print(
        f"Follow this link if it doesn't open automatically \n\n{request_uri}\n",
        file=sys.stderr,
    )
    webbrowser.open(request_uri)

    server.join()


def fetch_token(config) -> str:
    """Return cached API token if it's fresh enough.

    Otherwise, call login() and return the resulting token.
    """
    if config.api_token:
        return config.api_token

    login(config)
    return config.api_token



from __future__ import annotations

import abc

from collections.abc import Iterable
from pathlib import Path
from typing import Any, Protocol

import click
import yaml

from commodore.config import Config
from commodore.helpers import yaml_load_all


from .lint_dependency_specification import lint_components, lint_packages
from .lint_deprecated_parameters import lint_deprecated_parameters


class LintFunc(Protocol):
    def __call__(self, file: Path, filecontents: dict[str, Any]) -> int:
        ...


class Linter:
    @abc.abstractmethod
    def __call__(self, config: Config, path: Path) -> int:
        ...


class ComponentSpecLinter(Linter):
    def __call__(self, config: Config, path: Path) -> int:
        return run_linter(config, path, lint_components)


class DeprecatedParameterLinter(Linter):
    def __call__(self, config: Config, path: Path) -> int:
        return run_linter(config, path, lint_deprecated_parameters)


class PackageSpecLinter(Linter):
    def __call__(self, config: Config, path: Path) -> int:
        return run_linter(config, path, lint_packages)


LINTERS = {
    "components": ComponentSpecLinter(),
    "deprecated-parameters": DeprecatedParameterLinter(),
    "packages": PackageSpecLinter(),
}


def _lint_file(cfg: Config, file: Path, lintfunc: LintFunc) -> int:
    errcount = 0
    try:
        filecontents = yaml_load_all(file)
        if len(filecontents) == 0:
            if cfg.debug:
                click.echo(f"> Skipping empty file {file}")
        elif len(filecontents) > 1:
            if cfg.debug:
                click.echo(
                    f"> Skipping file {file}: Linting multi-document YAML streams is not supported",
                )
        elif not isinstance(filecontents[0], dict):
            if cfg.debug:
                click.echo(
                    f"> Skipping file {file}: Expected top-level dictionary in YAML document"
                )
        else:
            errcount = lintfunc(file, filecontents[0])

    except (yaml.YAMLError, UnicodeDecodeError) as e:
        if cfg.debug:
            click.echo(f"> Skipping file {file}: Unable to load as YAML: {e}")

    return errcount


def _lint_directory(cfg: Config, path: Path, lintfunc: LintFunc) -> int:
    if not path.is_dir():
        raise ValueError("Unexpected path argument: expected to be a directory")

    errcount = 0
    for dentry in path.iterdir():
        if dentry.stem.startswith("."):
            if cfg.debug:
                click.echo(f"> Skipping hidden directory entry {dentry}")
            continue
        if dentry.is_dir():
            errcount += _lint_directory(cfg, dentry, lintfunc)
        else:
            errcount += _lint_file(cfg, dentry, lintfunc)
    return errcount


def run_linter(cfg: Config, path: Path, lintfunc: LintFunc) -> int:
    """Run lint function `lintfunc` in `path`.

    If `path` is a directory, run the function in all `.ya?ml` files in the directory
    (recursively).
    If `path` is a file, run the lint function in that file, if it's a YAML file.

    Returns a value that can be used as exit code to indicate whether there were linting
    errors.
    """
    if path.is_dir():
        return _lint_directory(cfg, path, lintfunc)

    return _lint_file(cfg, path, lintfunc)


def check_removed_reclass_variables(
    config: Config, location: str, paths: Iterable[Path]
):
    lint = DeprecatedParameterLinter()

    errcount = 0
    for path in paths:
        errcount += lint(config, path)

    # Raise error if any linting errors occurred
    if errcount > 0:
        raise click.ClickException(
            f"Found {errcount} usages of removed reclass variables "
            + f"in the {location}. See individual lint errors for details."
        )



from __future__ import annotations

from collections.abc import Iterable
from pathlib import Path as P
from typing import Optional

import _jsonnet
import click

from commodore.gitrepo import GitRepo
from commodore.multi_dependency import MultiDependency


class Component:
    _name: str
    _repo: Optional[GitRepo]
    _dependency: MultiDependency
    _version: Optional[str] = None
    _dir: P
    _sub_path: str

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        name: str,
        dependency: MultiDependency,
        work_dir: Optional[P] = None,
        version: Optional[str] = None,
        directory: Optional[P] = None,
        sub_path: str = "",
    ):
        self._name = name
        if directory:
            self._dir = directory
        elif work_dir:
            self._dir = component_dir(work_dir, self.name)
        else:
            raise click.ClickException(
                "Either `work_dir` or `directory` must be provided."
            )
        self._dependency = dependency
        self._dependency.register_component(self.name, self._dir)
        self.version = version
        self._sub_path = sub_path
        self._repo = None

    @property
    def name(self) -> str:
        return self._name

    @property
    def repo(self) -> GitRepo:
        if not self._repo:
            self._repo = GitRepo(None, self._dir)
        return self._repo

    @property
    def dependency(self):
        return self._dependency

    @dependency.setter
    def dependency(self, dependency: MultiDependency):
        """Update the GitRepo backing the component"""
        self._dependency.deregister_component(self.name)
        dependency.register_component(self.name, self._dir)
        self._dependency = dependency
        # Clear worktree GitRepo wrapper when we update the component's backing
        # dependency. The new GitRepo wrapper will be created on the nex access of
        # `repo`.
        self._repo = None

    @property
    def repo_url(self) -> str:
        return self._dependency.url

    @property
    def version(self) -> Optional[str]:
        return self._version

    @version.setter
    def version(self, version: str):
        self._version = version

    @property
    def repo_directory(self) -> P:
        return self._dir

    @property
    def target_directory(self) -> P:
        return self._dir / self._sub_path

    @property
    def class_file(self) -> P:
        return self.target_directory / "class" / f"{self.name}.yml"

    @property
    def defaults_file(self) -> P:
        return self.target_directory / "class" / "defaults.yml"

    @property
    def lib_files(self) -> Iterable[P]:
        lib_dir = self.target_directory / "lib"
        if lib_dir.exists():
            for e in lib_dir.iterdir():
                # Skip hidden files in lib directory
                if not e.name.startswith("."):
                    yield e

        return []

    def get_library(self, libname: str) -> Optional[P]:
        lib_dir = self.target_directory / "lib"
        if not lib_dir.exists():
            return None

        for f in self.lib_files:
            if f.absolute() == P(lib_dir / libname).absolute():
                return f.absolute()

        return None

    @property
    def parameters_key(self):
        return component_parameters_key(self.name)

    def checkout(self):
        self._dependency.checkout_component(self.name, self.version)

    def render_jsonnetfile_json(self, component_params):
        """
        Render jsonnetfile.json from jsonnetfile.jsonnet
        """
        jsonnetfile_jsonnet = self._dir / "jsonnetfile.jsonnet"
        jsonnetfile_json = self._dir / "jsonnetfile.json"
        if jsonnetfile_jsonnet.is_file():
            if jsonnetfile_json.name in self.repo.repo.tree():
                click.secho(
                    f" > [WARN] Component {self.name} repo contains both jsonnetfile.json "
                    + "and jsonnetfile.jsonnet, continuing with jsonnetfile.jsonnet",
                    fg="yellow",
                )
            # pylint: disable=c-extension-no-member
            output = _jsonnet.evaluate_file(
                str(jsonnetfile_jsonnet),
                ext_vars=component_params.get("jsonnetfile_parameters", {}),
            )
            with open(self._dir / "jsonnetfile.json", "w", encoding="utf-8") as fp:
                fp.write(output)
                fp.write("\n")


def component_dir(work_dir: P, name: str) -> P:
    return work_dir / "dependencies" / name


def component_parameters_key(name: str) -> str:
    return name.replace("-", "_")



from __future__ import annotations

import datetime
import re
import tempfile
import shutil
import textwrap

from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Optional, Protocol, Sequence

import click

from commodore.config import Config
from commodore.gitrepo import GitRepo
from commodore.multi_dependency import MultiDependency

SLUG_REGEX = re.compile("^[a-z][a-z0-9-]+[a-z0-9]$")


class Renderer(Protocol):
    def __call__(
        self,
        template_location: str,
        extra_context: dict[str, Any],
        no_input: bool,
        output_dir: Path,
    ):
        ...


class Templater(ABC):
    config: Config
    _slug: str
    _name: Optional[str]
    github_owner: str
    copyright_holder: str
    golden_tests: bool
    today: datetime.date
    output_dir: Optional[Path] = None

    def __init__(
        self,
        config: Config,
        slug: str,
        name: Optional[str] = None,
        output_dir: str = "",
    ):
        self.config = config
        self.slug = slug
        self._name = name
        self.today = datetime.date.today()
        if output_dir != "":
            odir = Path(output_dir)
            if not odir.is_dir():
                raise click.ClickException(f"Output directory {odir} doesn't exist")

            self.output_dir = odir

    def _validate_slug(self, value: str) -> str:
        if value.startswith(f"{self.deptype}-"):
            raise click.ClickException(
                f"The {self.deptype} slug may not start with '{self.deptype}-'"
            )
        if not SLUG_REGEX.match(value):
            raise click.ClickException(
                f"The {self.deptype} slug must match '{SLUG_REGEX.pattern}'"
            )
        return value

    @property
    def slug(self) -> str:
        return self._slug

    @slug.setter
    def slug(self, value: str):
        self._slug = self._validate_slug(value)

    @property
    def name(self) -> str:
        if not self._name:
            return self.slug
        return self._name

    @property
    def repo_url(self) -> str:
        return f"git@github.com:{self.github_owner}/{self.deptype}-{self.slug}.git"

    @property
    @abstractmethod
    def deptype(self) -> str:
        """Return dependency type of template as string.

        The base implementation of `_validate_slug()` will reject slugs which are
        prefixed with the value of this property.
        """

    @property
    @abstractmethod
    def target_dir(self) -> Path:
        """Return Path indicating where to render the template to."""

    @property
    @abstractmethod
    def template(self) -> str:
        """Path or URL of the template to render"""

    @property
    @abstractmethod
    def template_renderer(self) -> Renderer:
        """Template rendering function to use for the template.

        Allows child classes to select either plain cookiecutter or cruft.
        """

    @property
    @abstractmethod
    def cookiecutter_args(self) -> dict[str, str]:
        """Cookiecutter template inputs.

        Passed to the rendering function as `extra_context`
        """

    @property
    @abstractmethod
    def additional_files(self) -> Sequence[str]:
        """Sequence of additional files to include in the initial Git commit."""

    def create(self) -> None:
        click.secho(f"Adding {self.deptype} {self.name}...", bold=True)

        if self.target_dir.exists():
            raise click.ClickException(
                f"Unable to add {self.deptype} {self.name}: "
                + f"{self.target_dir} already exists."
            )

        want_worktree = (
            self.config.inventory.dependencies_dir in self.target_dir.parents
        )
        if want_worktree:
            md = MultiDependency(self.repo_url, self.config.inventory.dependencies_dir)
            md.initialize_worktree(self.target_dir)

        with tempfile.TemporaryDirectory() as tmpdir:
            self.template_renderer(
                self.template,
                no_input=True,
                output_dir=Path(tmpdir),
                extra_context=self.cookiecutter_args,
            )
            shutil.copytree(
                Path(tmpdir) / self.slug, self.target_dir, dirs_exist_ok=True
            )

        self.commit("Initial commit", amend=want_worktree)
        click.secho(
            f"{self.deptype.capitalize()} {self.name} successfully added 🎉", bold=True
        )

    def commit(self, msg: str, amend=False, init=True) -> bool:
        # If we're amending an existing commit, we don't want to force initialize the
        # repo.
        repo = GitRepo(self.repo_url, self.target_dir, force_init=not amend and init)

        # stage_all() returns the full diff compared to the last commit. Therefore, we
        # do stage_files() first and then stage_all(), to ensure we get the complete
        # diff.
        repo.stage_files(self.additional_files)
        diff_text, changed = repo.stage_all()

        if changed:
            indented = textwrap.indent(diff_text, "     ")
            message = f" > Changes:\n{indented}"
        else:
            message = " > No changes."
        click.echo(message)

        if changed:
            # Only create a new commit if there are any changes.
            repo.commit(msg, amend=amend)
        return changed



from __future__ import annotations

import difflib
import shutil
import time
import textwrap

from collections.abc import Iterable
from pathlib import Path as P

import click
import yaml

from .component import Component
from .gitrepo import GitRepo, GitCommandError
from .helpers import ApiError, rm_tree_contents, lieutenant_query, sliding_window
from .cluster import Cluster
from .config import Config, Migration
from .k8sobject import K8sObject


def fetch_catalog(config: Config, cluster: Cluster) -> GitRepo:
    click.secho("Updating cluster catalog...", bold=True)
    repo_url = cluster.catalog_repo_url
    if config.debug:
        click.echo(f" > Cloning cluster catalog {repo_url}")
    return GitRepo.clone(repo_url, config.catalog_dir, config)


def _pretty_print_component_commit(name, component: Component) -> str:
    short_sha = component.repo.head_short_sha
    return f" * {name}: {component.version} ({short_sha})"


def _pretty_print_config_commit(name, repo: GitRepo) -> str:
    short_sha = repo.head_short_sha
    return f" * {name}: {short_sha}"


def _render_catalog_commit_msg(cfg) -> str:
    # pylint: disable=import-outside-toplevel
    import datetime

    now = datetime.datetime.now().isoformat(timespec="milliseconds")

    component_commits = [
        _pretty_print_component_commit(cn, c) for cn, c in cfg.get_components().items()
    ]
    component_commits_str = "\n".join(component_commits)

    config_commits = [
        _pretty_print_config_commit(c, r) for c, r in cfg.get_configs().items()
    ]
    config_commits_str = "\n".join(config_commits)

    return f"""Automated catalog update from Commodore

Component commits:
{component_commits_str}

Configuration commits:
{config_commits_str}

Compilation timestamp: {now}
"""


def clean_catalog(repo: GitRepo):
    if repo.working_tree_dir is None:
        raise click.ClickException("Catalog repo has no working tree")

    catalogdir = P(repo.working_tree_dir, "manifests")
    click.secho("Cleaning catalog repository...", bold=True)
    # delete everything in catalog
    if catalogdir.is_dir():
        rm_tree_contents(catalogdir)
    else:
        click.echo(
            f" > Skipping cleaning, catalog directory {catalogdir} doesn't exist"
        )


def _push_catalog(cfg: Config, repo: GitRepo, commit_message: str):
    """Push catalog to catalog repo if conditions to allow push are met.

    Conditions to allow pushing are:
    * Commodore doesn't run in local mode
    * User has requested pushing with `--push`

    Ask user to confirm push if `--interactive` is specified
    """
    if cfg.local:
        repo.reset(working_tree=False)
        click.echo(" > Skipping commit+push to catalog in local mode...")
        return

    if cfg.interactive and cfg.push:
        cfg.push = click.confirm(" > Should the push be done?")

    if cfg.push:
        click.echo(" > Commiting changes...")
        repo.commit(commit_message)
        click.echo(" > Pushing catalog to remote...")
        try:
            pushinfos = repo.push()
        except GitCommandError as e:
            raise click.ClickException(
                "Failed to push to the catalog repository: "
                + f"Git exited with status code {e.status}"
            ) from e
        for pi in pushinfos:
            # Any error has pi.ERROR set in the `flags` bitmask
            # We just forward the summary from the first pushinfo which has the error
            # flag set.
            summary = pi.summary.strip()
            if (pi.flags & pi.ERROR) != 0:
                raise click.ClickException(
                    f"Failed to push to the catalog repository: {summary}"
                )
    else:
        click.echo(" > Skipping commit+push to catalog...")
        click.echo(" > Use flag --push to commit and push the catalog repo")
        click.echo(" > Add flag --interactive to show the diff and decide on the push")


def _is_semantic_diff_kapitan_029_030(win: tuple[str, str]) -> bool:
    """
    Returns True if a pair of lines of a diff which is already sorted
    by K8s object indicates that this diff contains a semantic change
    when migrating from  Kapitan 0.29 to 0.30.

    The function expects pairs of diff lines as input.

    The function treats the following diffs as non-semantic:
    * Change of "app.kubernetes.io/managed-by: Tiller" to
      "app.kubernetes.io/managed-by: Helm"
    * Change of "heritage: Tiller" to "heritage: Helm"
    * `null` objects not emitted in multi-object YAML documents anymore
    """
    line_a, line_b = map(str.rstrip, win)

    # Ignore context and metadata lines:
    if (
        # We don't use line_a/line_b here, as it's possible that the leading space of a
        # context line gets stripped if the line is empty.
        win[0].startswith(" ")
        or win[1].startswith(" ")
        or line_a.startswith("@@")
        or line_b.startswith("@@")
    ):
        return False

    # Ignore changes where we don't emit a null object preceded or followed
    # by a stream separator anymore
    if line_a == "-null" and line_b in ("----", "---- null"):
        return False
    if line_a == "---- null" and line_b in ("----", "---- null"):
        return False

    # Ignore changes which are only about Tiller -> Helm as object manager
    if line_a.startswith("-") and line_b.startswith("+"):
        if line_a.endswith("app.kubernetes.io/managed-by: Tiller") and line_b.endswith(
            "app.kubernetes.io/managed-by: Helm"
        ):
            return False
        if line_a.endswith("heritage: Tiller") and line_b.endswith("heritage: Helm"):
            return False

    # Don't ignore any other diffs
    return True


def _kapitan_029_030_difffunc(
    before_text: str, after_text: str, fromfile: str = "", tofile: str = ""
) -> tuple[Iterable[str], bool]:

    before_objs = sorted(yaml.safe_load_all(before_text), key=K8sObject)
    before_sorted_lines = yaml.dump_all(before_objs).split("\n")

    after_objs = sorted(yaml.safe_load_all(after_text), key=K8sObject)
    after_sorted_lines = yaml.dump_all(after_objs).split("\n")

    diff = difflib.unified_diff(
        before_sorted_lines,
        after_sorted_lines,
        lineterm="",
        fromfile=fromfile,
        tofile=tofile,
    )
    diff_lines = list(diff)
    suppress_diff = not any(
        _is_semantic_diff_kapitan_029_030(win)
        for win in sliding_window(diff_lines[2:], 2)
    )

    return diff_lines, suppress_diff


def update_catalog(cfg: Config, targets: Iterable[str], repo: GitRepo):
    if repo.working_tree_dir is None:
        raise click.ClickException("Catalog repo has no working tree")

    click.secho("Updating catalog repository...", bold=True)

    catalogdir = P(repo.working_tree_dir, "manifests")
    for target_name in targets:
        shutil.copytree(
            cfg.inventory.output_dir / target_name, catalogdir, dirs_exist_ok=True
        )

    start = time.time()
    if cfg.migration == Migration.KAP_029_030:
        click.echo(" > Smart diffing started... (this can take a while)")
        difftext, changed = repo.stage_all(diff_func=_kapitan_029_030_difffunc)
    else:
        difftext, changed = repo.stage_all()
    elapsed = time.time() - start

    if changed:
        indented = textwrap.indent(difftext, "     ")
        message = f" > Changes:\n{indented}"
        if cfg.migration:
            message += f"\n > Smart diffing took {elapsed:.2f}s"
    else:
        message = " > No changes."
    click.echo(message)

    commit_message = _render_catalog_commit_msg(cfg)
    if cfg.debug:
        click.echo(" > Commit message will be")
        click.echo(textwrap.indent(commit_message, "   "))
    if changed:
        _push_catalog(cfg, repo, commit_message)
    else:
        click.echo(" > Skipping commit+push to catalog...")


def catalog_list(cfg):
    try:
        clusters = lieutenant_query(cfg.api_url, cfg.api_token, "clusters", "")
    except ApiError as e:
        raise click.ClickException(f"While listing clusters on Lieutenant: {e}") from e
    for cluster in clusters:
        display_name = cluster["displayName"]
        catalog_id = cluster["id"]
        if cfg.verbose:
            click.secho(catalog_id, nl=False, bold=True)
            click.echo(f" - {display_name}")
        else:
            click.echo(catalog_id)



from __future__ import annotations

import json
import os

from pathlib import Path
from typing import Any

import click
from xdg.BaseDirectory import xdg_cache_home

cache_name = Path(xdg_cache_home) / "commodore" / "token"


def save(lieutenant: str, token: dict[str, Any]):
    try:
        with open(cache_name, "r", encoding="utf-8") as f:
            cache = json.load(f)
    except (IOError, FileNotFoundError):
        cache = {}
    except json.JSONDecodeError:
        click.secho(" > Dropping invalid JSON for token cache", fg="yellow")
        cache = {}

    cache[lieutenant] = token

    os.makedirs(os.path.dirname(cache_name), exist_ok=True)
    with open(cache_name, "w", encoding="utf-8") as f:
        f.write(json.dumps(cache, indent=1))


def get(lieutenant: str) -> dict[str, Any]:
    try:
        with open(cache_name, "r", encoding="utf-8") as f:
            cache = json.load(f)
    except (IOError, FileNotFoundError, json.JSONDecodeError):
        return {}
    data = cache.get(lieutenant, {})
    if isinstance(data, str):
        data = {}
    return data



from __future__ import annotations

from collections.abc import Iterable
from pathlib import Path
from typing import Any

import click

from .catalog import fetch_catalog, clean_catalog, update_catalog
from .cluster import (
    Cluster,
    load_cluster_from_api,
    read_cluster_and_tenant,
    update_params,
    update_target,
)
from .config import Config
from .dependency_mgmt import (
    fetch_components,
    fetch_packages,
    register_components,
    register_packages,
    verify_version_overrides,
)
from .dependency_mgmt.component_library import create_component_library_aliases
from .dependency_mgmt.jsonnet_bundler import (
    fetch_jsonnet_libraries,
    jsonnet_dependencies,
)
from .gitrepo import GitRepo
from .helpers import (
    ApiError,
    clean_working_tree,
    kapitan_compile,
    kapitan_inventory,
    rm_tree_contents,
)
from .inventory.lint import check_removed_reclass_variables
from .postprocess import postprocess_components
from .refs import update_refs


def check_removed_reclass_variables_inventory(config: Config, tenant: str):
    paths = [
        # global defaults
        config.inventory.global_config_dir,
        # tenant repo
        config.inventory.tenant_config_dir(tenant),
    ]

    check_removed_reclass_variables(config, "hierarchy", paths)


def check_removed_reclass_variables_components(config: Config):
    paths: Iterable[Path] = sum(
        # defaults.yml and <component-name>.yml for all enabled components
        [[c.defaults_file, c.class_file] for c in config.get_components().values()],
        [],
    )

    check_removed_reclass_variables(config, "enabled components", paths)


def _fetch_global_config(cfg: Config, cluster: Cluster):
    click.secho("Updating global config...", bold=True)
    repo = GitRepo(cluster.global_git_repo_url, cfg.inventory.global_config_dir)
    rev = cluster.global_git_repo_revision
    if cfg.global_repo_revision_override:
        rev = cfg.global_repo_revision_override
    repo.checkout(rev)
    cfg.register_config("global", repo)


def _fetch_customer_config(cfg: Config, cluster: Cluster):
    click.secho("Updating customer config...", bold=True)
    repo_url = cluster.config_repo_url
    if cfg.debug:
        click.echo(f" > Cloning customer config {repo_url}")
    repo = GitRepo(repo_url, cfg.inventory.tenant_config_dir(cluster.tenant_id))
    rev = cluster.config_git_repo_revision
    if cfg.tenant_repo_revision_override:
        rev = cfg.tenant_repo_revision_override
    repo.checkout(rev)
    cfg.register_config("customer", repo)


def _regular_setup(config: Config, cluster_id):
    try:
        cluster = load_cluster_from_api(config, cluster_id)
    except ApiError as e:
        raise click.ClickException(f"While fetching cluster specification: {e}") from e

    update_target(config, config.inventory.bootstrap_target)
    update_params(config.inventory, cluster)

    # Fetch components and config
    _fetch_global_config(config, cluster)
    _fetch_customer_config(config, cluster)
    check_removed_reclass_variables_inventory(config, cluster.tenant_id)

    # Fetch component config packages. This needs to happen before component fetching
    # because we want to be able to discover components included by config packages.
    fetch_packages(config)

    fetch_components(config)

    update_target(config, config.inventory.bootstrap_target)

    for alias, component in config.get_component_aliases().items():
        update_target(config, alias, component=component)

    # Fetch catalog
    return fetch_catalog(config, cluster)


def _local_setup(config: Config, cluster_id):
    click.secho("Running in local mode", bold=True)
    click.echo(" > Will use existing inventory, components, and catalog")
    if not config.fetch_dependencies:
        click.echo(" > Will use existing Jsonnet and Kapitan dependencies")
        click.echo(
            "   > Use --fetch-dependencies at least once if you're trying to enable "
            + "a new component in local mode, otherwise Kapitan will fail to find the component"
        )

    file = config.inventory.target_file(config.inventory.bootstrap_target)
    if not file.is_file():
        raise click.ClickException(f"Invalid working dir state: '{file}' is missing")

    click.echo(" > Assert current working dir state matches requested compilation")
    current_cluster_id, tenant = read_cluster_and_tenant(config.inventory)
    if current_cluster_id != cluster_id:
        error = (
            "[Local mode] Cluster ID mismatch: local state targets "
            + f"{current_cluster_id}, compilation was requested for {cluster_id}"
        )
        raise click.ClickException(error)

    click.secho("Registering config...", bold=True)
    config.register_config("global", GitRepo(None, config.inventory.global_config_dir))
    config.register_config(
        "customer", GitRepo(None, config.inventory.tenant_config_dir(tenant))
    )

    check_removed_reclass_variables_inventory(config, tenant)

    click.secho("Resetting targets...", bold=True)
    rm_tree_contents(config.inventory.targets_dir)

    click.secho("Resetting component symlinks in inventory...", bold=True)
    rm_tree_contents(config.inventory.defaults_dir)
    rm_tree_contents(config.inventory.components_dir)

    click.secho("Creating bootstrap target...", bold=True)
    update_target(config, config.inventory.bootstrap_target)

    register_packages(config)
    register_components(config)

    for alias, component in config.get_component_aliases().items():
        update_target(config, alias, component=component)
    update_target(config, config.inventory.bootstrap_target)

    click.secho("Configuring catalog repo...", bold=True)
    return GitRepo(None, config.catalog_dir)


def check_parameters_component_versions(cluster_parameters):
    """
    Check inventory for `parameters.component_versions`.

    Raise an error if the parameter has any contents.
    """
    cvers = cluster_parameters.get("component_versions", {})
    if len(cvers.keys()) > 0:
        raise click.ClickException(
            "Specifying component versions in parameter `component_versions` "
            + "is no longer suppported. Please migrate your configuration to "
            + "parameter `components`."
        )


def setup_compile_environment(config: Config) -> tuple[dict[str, Any], Iterable[str]]:
    # Raise error if any enabled components use removed reclass variables
    check_removed_reclass_variables_components(config)

    inventory = kapitan_inventory(config)
    cluster_parameters = inventory[config.inventory.bootstrap_target]["parameters"]
    check_parameters_component_versions(cluster_parameters)
    create_component_library_aliases(config, cluster_parameters)

    # Verify that all aliased components support instantiation
    config.verify_component_aliases(cluster_parameters)
    config.register_component_deprecations(cluster_parameters)
    # Raise exception if component version override without URL is present in the
    # hierarchy.
    verify_version_overrides(cluster_parameters)

    for component in config.get_components().values():
        ckey = component.parameters_key
        component.render_jsonnetfile_json(cluster_parameters[ckey])

    if config.fetch_dependencies:
        fetch_jsonnet_libraries(config.work_dir, deps=jsonnet_dependencies(config))

    aliases = config.get_component_aliases()

    # Generate Kapitan secret references from refs found in inventory
    # parameters
    update_refs(config, aliases, inventory)

    return inventory, list(aliases.keys())


# pylint: disable=redefined-builtin
def compile(config, cluster_id):
    if config.local:
        catalog_repo = _local_setup(config, cluster_id)
    else:
        clean_working_tree(config)
        catalog_repo = _regular_setup(config, cluster_id)

    inventory, targets = setup_compile_environment(config)

    clean_catalog(catalog_repo)

    kapitan_compile(config, targets, search_paths=[config.vendor_dir])

    postprocess_components(config, inventory, config.get_components())

    update_catalog(config, targets, catalog_repo)

    click.secho("Catalog compiled! 🎉", bold=True)

    config.print_deprecation_notices()



from __future__ import annotations

import difflib
import hashlib
import shutil

from collections import namedtuple
from collections.abc import Iterable, Sequence
from pathlib import Path
from typing import Optional, Protocol

import click

from git import Actor, BadName, FetchInfo, GitCommandError, PushInfo, Repo
from git.objects import Tree

from url_normalize import url_normalize
from url_normalize.tools import deconstruct_url, reconstruct_url


class RefError(ValueError):
    pass


CommitInfo = namedtuple("CommitInfo", ["commit", "branch", "tag"])


class DiffFunc(Protocol):
    def __call__(
        self, before_text: str, after_text: str, fromfile: str = "", tofile: str = ""
    ) -> tuple[Iterable[str], bool]:
        ...


def _normalize_git_ssh(url: str) -> str:
    # Import url_normalize internal methods here, so they're not visible in the file
    # scope of gitrepo.py
    # pylint: disable=import-outside-toplevel
    from url_normalize.url_normalize import (
        normalize_userinfo,
        normalize_host,
        normalize_path,
        provide_url_scheme,
    )

    if "@" in url and not url.startswith("ssh://"):
        # Assume git@host:repo format, reformat so url_normalize understands
        # the URL
        host, repo = url.split(":")
        url = f"{host}/{repo}"
    # Reuse normalization logic from url_normalize, simplify for Git-SSH use case.
    # We can't do `url_normalize(url, "ssh"), because the library doesn't know "ssh" as
    # a scheme, and fails to look up the default port for "ssh".
    url = provide_url_scheme(url, "ssh")
    urlparts = deconstruct_url(url)
    urlparts = urlparts._replace(
        userinfo=normalize_userinfo(urlparts.userinfo),
        host=normalize_host(urlparts.host),
        path=normalize_path(urlparts.path, scheme="https"),
    )
    return reconstruct_url(urlparts)


def normalize_git_url(url: str) -> str:
    """Normalize HTTP(s) and SSH Git URLs"""
    if "@" in url and ("://" not in url or url.startswith("ssh://")):
        url = _normalize_git_ssh(url)
    elif url.startswith("http://") or url.startswith("https://"):
        url = url_normalize(url)
    return url


def _colorize_diff(line: str) -> str:
    if line.startswith("--- ") or line.startswith("+++ ") or line.startswith("@@ "):
        return click.style(line, fg="yellow")
    if line.startswith("+"):
        return click.style(line, fg="green")
    if line.startswith("-"):
        return click.style(line, fg="red")
    return line


def _compute_similarity(change):
    before = change.b_blob.data_stream.read().decode("utf-8").split("\n")
    after = change.a_blob.data_stream.read().decode("utf-8").split("\n")
    r = difflib.SequenceMatcher(a=before, b=after).ratio()
    similarity_diff = []
    similarity_diff.append(click.style(f"--- {change.b_path}", fg="yellow"))
    similarity_diff.append(click.style(f"+++ {change.a_path}", fg="yellow"))
    similarity_diff.append(f"Renamed file, similarity index {r * 100:.2f}%")
    return similarity_diff


def _default_difffunc(
    before_text: str, after_text: str, fromfile: str = "", tofile: str = ""
) -> tuple[Iterable[str], bool]:
    before_lines = before_text.split("\n")
    after_lines = after_text.split("\n")
    diff_lines = difflib.unified_diff(
        before_lines, after_lines, lineterm="", fromfile=fromfile, tofile=tofile
    )
    # never suppress diffs in default difffunc
    return diff_lines, False


def _process_diff(change_type: str, change, diff_func: DiffFunc) -> Iterable[str]:
    difftext = []
    # Because we're diffing the staged changes, the diff objects
    # are backwards, and "added" files are actually being deleted
    # and vice versa for "deleted" files.
    if change_type == "A":
        difftext.append(click.style(f"Deleted file {change.b_path}", fg="red"))
    elif change_type == "D":
        difftext.append(click.style(f"Added file {change.b_path}", fg="green"))
    elif change_type == "R":
        difftext.append(
            click.style(f"Renamed file {change.b_path} => {change.a_path}", fg="yellow")
        )
    else:
        # Other changes should produce a usable diff
        # The diff objects are backwards, so use b_blob as before
        # and a_blob as after.
        before = change.b_blob.data_stream.read().decode("utf-8")
        after = change.a_blob.data_stream.read().decode("utf-8")
        diff_lines, suppress_diff = diff_func(
            before, after, fromfile=change.b_path, tofile=change.a_path
        )
        if not suppress_diff:
            if change.renamed_file:
                # Just compute similarity ratio for renamed files
                # similar to git's diffing
                difftext.append("\n".join(_compute_similarity(change)).strip())
            else:
                diff_lines = [_colorize_diff(line) for line in diff_lines]
                difftext.append("\n".join(diff_lines).strip())

    return difftext


class GitRepo:
    _repo: Repo

    @classmethod
    def clone(cls, repository_url: str, directory: Path, cfg):
        """Clone repository.

        Create initial commit on master branch, if remote is empty."""
        name = None
        email = None
        if cfg:
            name = cfg.username
            email = cfg.usermail
        r = GitRepo(
            repository_url,
            directory,
            force_init=True,
            author_name=name,
            author_email=email,
        )
        try:
            r.checkout()
        except RefError as e:
            click.echo(f" > {e}, creating initial commit for {directory}")
            r.commit("Initial commit")
        except Exception as e:
            raise click.ClickException(
                f"While cloning git repository from {repository_url}: {e}"
            ) from e
        return r

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        remote: Optional[str],
        targetdir: Path,
        force_init=False,
        author_name: Optional[str] = None,
        author_email: Optional[str] = None,
        config=None,
        bare=False,
    ):
        if not force_init and targetdir.exists():
            self._repo = Repo(targetdir)
        else:
            self._repo = Repo.init(targetdir, bare=bare)

        if remote:
            self.remote = remote

        if author_name and author_email:
            self._author = Actor(author_name, author_email)
        else:
            self._author = Actor.committer(self._repo.config_reader())

        if config:
            self._debug = config.debug
            self._trace = config.trace
        else:
            self._debug = False
            self._trace = False

    @property
    def repo(self) -> Repo:
        return self._repo

    @property
    def remote(self) -> str:
        return self._repo.remote().url

    @remote.setter
    def remote(self, remote: str):
        remote = normalize_git_url(remote)
        try:
            self._repo.remote().set_url(remote)
        except ValueError:
            self._repo.create_remote("origin", remote)

        # Generate a best effort push-over-SSH URL for http(s) repo URLs.
        if remote.startswith(("http://", "https://")):
            remote_parts = deconstruct_url(remote)
            # Ignore everything but the host and path parts of the URL to
            # build the push URL
            pushurl = f"ssh://git@{remote_parts.host}{remote_parts.path}"
            self._repo.remote().set_url(pushurl, push=True)

    @property
    def working_tree_dir(self) -> Optional[Path]:
        d = self._repo.working_tree_dir
        if d:
            return Path(d)
        return None

    @property
    def head_short_sha(self) -> str:
        sha = self._repo.head.commit.hexsha
        return self._repo.git.rev_parse(sha, short=6)

    @property
    def _author_env(self) -> dict[str, str]:
        return {
            Actor.env_author_name: self._author.name or "",
            Actor.env_author_email: self._author.email or "",
            Actor.env_committer_name: self._author.name or "",
            Actor.env_committer_email: self._author.email or "",
        }

    @property
    def _null_tree(self) -> Tree:
        """Generate empty Tree for the repo.
        An empty Git tree is represented by the C string "tree 0".
        The hash of the empty tree is always SHA1("tree 0\\0").  This method computes
        the hexdigest of this sha1 and creates a tree object for the empty tree of
        `self._repo`.
        """
        null_tree_sha = hashlib.sha1(b"tree 0\0").hexdigest()  # nosec
        return self._repo.tree(null_tree_sha)

    def _remote_prefix(self) -> str:
        """
        Find prefix of Git remote, will usually be 'origin/'.
        """
        return self._repo.remote().name + "/"

    def _default_version(self) -> str:
        """
        Find default branch of the remote
        """
        try:
            version = self._repo.remote().refs["HEAD"].reference.name
        except IndexError:
            try:
                self._repo.git.remote("set-head", "origin", "--auto")
                version = self._repo.remote().refs["HEAD"].reference.name
            except GitCommandError:
                # If we don't have a remote HEAD, we fall back to creating a master
                # branch.
                version = f"{self._remote_prefix()}master"
        return version.replace(self._remote_prefix(), "", 1)

    def _find_commit_for_version(
        self, version: str, remote_heads: Iterable[FetchInfo]
    ) -> CommitInfo:
        remote_prefix = self._remote_prefix()
        for head in remote_heads:
            tag = None
            branch = None
            headname = head.name
            if headname.startswith(remote_prefix):
                branch = headname.replace(remote_prefix, "", 1)
                headname = branch
            else:
                tag = headname

            if headname == version:
                commit = head.commit
                break
        else:
            # If we haven't found a branch or tag matching the requested version,
            # assume the version is a commit sha.
            commit = self._repo.rev_parse(version)
            branch = None
            tag = None

        return CommitInfo(commit=commit, branch=branch, tag=tag)

    def fetch(
        self, remote: str = "origin", tags: bool = True, prune: bool = True
    ) -> Iterable[FetchInfo]:
        return self._repo.remote(remote).fetch(tags=tags, prune=prune)

    def has_local_branches(self) -> bool:
        local_heads = set(h.name for h in self.repo.heads)
        remote_prefix = self._remote_prefix()
        remote_heads = set(h.name.replace(remote_prefix, "", 1) for h in self.fetch())
        return len(local_heads - remote_heads) > 0

    def has_local_changes(self) -> bool:
        return self._repo.is_dirty() or len(self._repo.untracked_files) > 0

    def _create_worktree(self, worktree: Path, version: str):
        """Create worktree.

        This method expects `worktree` to not exist."""

        # We need to use `git.execute()` for the worktree commands as GitPython only has
        # basic support for worktrees.
        self._repo.git.execute(["git", "worktree", "prune"])
        cmd = ["git", "worktree", "add", "-f", str(worktree), version]
        try:
            self._repo.git.execute(cmd)
        except GitCommandError as e:
            # Assume that GitCommandError is only caused by invalid versions
            raise RefError(f"Failed to checkout revision '{version}'") from e

    def _migrate_to_worktree(self, wtr: GitRepo, worktree: Path, version: str):
        """Migrate non-worktree checkout to worktree."""
        if wtr.has_local_branches() or wtr.has_local_changes():
            raise click.ClickException(
                f"Migrating dependency {worktree} to worktree-based checkout "
                + "would delete uncommitted changes, untracked files, or "
                + "unpushed branches, aborting..."
            )
        click.secho(f" > Removing non-worktree based checkout {worktree}", fg="yellow")
        shutil.rmtree(worktree)
        self._create_worktree(worktree, version)

    def _update_worktree_remote(self, wtr: GitRepo, worktree: Path, version: str):
        """Update existing worktree checkout to new remote.

        Updating the remote for a worktree needs special handling, since we generally
        don't want to update the remote URL for the previously used remote URL's bare
        clone, but instead recreate the worktree from the new remote's bare clone."""
        if wtr.has_local_changes():
            raise click.ClickException(
                f"Switching remote for worktree '{worktree}' would delete "
                + "uncommitted changes or untracked files, aborting..."
            )
        # Remove stale worktree if there are no uncommitted changes
        wtr_common_dir = Path(wtr.repo.common_dir).resolve()
        click.secho(
            f" > Removing stale, but clean worktree {worktree}, "
            + f"bare repo is available at {wtr_common_dir}",
            fg="green",
        )
        wtr.repo.git.execute(["git", "worktree", "remove", str(worktree)])
        self._create_worktree(worktree, version)

    def _checkout_existing_worktree(self, worktree: Path, version: str):
        """Perform checkout if requested worktree directory already exists.

        The heavy work is generally done by `_migrate_to_worktree()`,
        `_update_worktree_remote()` or `checkout()`.
        """
        wtr = GitRepo(None, worktree)
        if not wtr.repo.has_separate_working_tree():
            # If the worktree's common dir is stored in the repository working tree
            # root, we're migrating from a non-worktree checkout to a worktree checkout.
            self._migrate_to_worktree(wtr, worktree, version)
        elif wtr.remote != self.remote:
            # If the existing directory is already a worktree, but we're using a
            # different remote for the requested worktree, we need to recreate the
            # worktree from the new remote's bare clone.
            self._update_worktree_remote(wtr, worktree, version)
        else:
            # Otherwise, we just need to update the worktree's version. We simply use
            # `checkout()` in the worktree to do so.
            wtr.checkout(version)

    def checkout_worktree(self, worktree: Path, version: Optional[str]):
        """Create worktree if it doesn't exist and check out `version` in it.

        If `version` is not provided, the remote's default branch is checked out.

        If a repo which isn't a worktree of `self` is found in the requested worktree
        location, the method will try to replace the old checkout with the requested
        worktree unless there's any local changes (untracked files, uncommitted changes,
        or branches which don't exist upstream).
        """
        # Try to fetch remote heads, so we can actually check them out
        try:
            _ = self.fetch()
        except ValueError:
            pass

        if version is None:
            version = self._default_version()

        # If the worktree directory exists, use `_checkout_existing_worktree()`
        if worktree.is_dir():
            self._checkout_existing_worktree(worktree, version)
            return

        # If the worktree directory doesn't exist yet, create the worktree
        self._create_worktree(worktree, version)

    def initialize_worktree(
        self, worktree: Path, initial_branch: Optional[str] = None
    ) -> None:
        if not initial_branch:
            initial_branch = self._default_version()

        # We need an initial commit to be able to create a worktree. Create initial
        # commit from empty tree.
        initsha = self._repo.git.execute(  # type: ignore[call-overload]
            command=[
                "git",
                "commit-tree",
                "-m",
                "Initial commit",
                self._null_tree.hexsha,
            ],
            env=self._author_env,
        )

        # Create worktree using the provided branch name
        self._repo.git.execute(
            ["git", "worktree", "add", str(worktree), initsha, "-b", initial_branch]
        )

    @property
    def worktrees(self) -> list[GitRepo]:
        """List all worktrees for the repo"""
        # First prune worktrees, to ensure repo worktree state is clean
        self._repo.git.execute(["git", "worktree", "prune"])
        worktrees: list[GitRepo] = []
        wt_list = self._repo.git.execute(
            ["git", "worktree", "list", "--porcelain"],
            as_process=False,
            with_extended_output=False,
            stdout_as_string=True,
        ).splitlines()
        for line in wt_list:
            if " " not in line:
                continue
            k, v = line.split(" ")
            if k == "worktree":
                worktrees.append(GitRepo(None, Path(v)))

        return worktrees

    def checkout(self, version: Optional[str] = None):
        remote_heads = self.fetch()
        if not remote_heads:
            # GitPython's fetch-info parsing chokes on lines like
            # "   (refs/remotes/origin/HEAD has become dangling)"
            # see also https://github.com/gitpython-developers/GitPython/issues/962.
            # We handle this case by simply performing a second fetch which should
            # return fetch-infos with flags = 4 (HEAD_UPTODATE).
            remote_heads = self.fetch()

        if version is None:
            # Handle case where we want the default branch of the remote
            version = self._default_version()

        try:
            commit, branch, tag = self._find_commit_for_version(version, remote_heads)
            if branch:
                # If we found a remote branch for the requested version, find
                # or create a local branch and point HEAD to the branch.
                _head = [h for h in self._repo.heads if h.name == branch]
                if len(_head) > 0:
                    head = _head[0]
                    head.commit = commit
                else:
                    head = self._repo.create_head(branch, commit=commit)

                head.set_tracking_branch(self.repo.remote().refs[branch])
                self._repo.head.set_reference(head)
            elif tag:
                # Simply create detached head pointing to tag
                self._repo.head.set_reference(tag)
            else:
                # Create detached head by setting repo.head.reference as
                # direct ref to commit object.
                self._repo.head.set_reference(commit)

            # Reset working tree to current HEAD reference
            self._repo.head.reset(index=True, working_tree=True)
        except GitCommandError as e:
            raise RefError(f"Failed to checkout revision '{version}'") from e
        except BadName as e:
            raise RefError(f"Revision '{version}' not found in repository") from e

    def stage_all(self, diff_func: DiffFunc = _default_difffunc) -> tuple[str, bool]:
        """Stage all changes.
        This method currently doesn't handle hidden files correctly.

        This method returns a tuple containing the colorized diff of the staged changes
        and a boolean indicating whether any changes were staged.
        """
        index = self._repo.index

        # Stage deletions
        dels = index.diff(None)
        if dels:
            to_remove = []
            for c in dels.iter_change_type("D"):
                to_remove.append(c.b_path)
            if len(to_remove) > 0:
                index.remove(items=to_remove)

        # Stage all remaining changes
        index.add("*")
        # Compute diff of all changes
        try:
            diff = index.diff(self._repo.head.commit)
        except ValueError:
            # Assume that we're in an empty repo if we get a ValueError from
            # index.diff(repo.head.commit). Diff against empty tree.
            diff = index.diff(self._null_tree)

        changed = False
        difftext: list[str] = []
        if diff:
            changed = True
            for ct in diff.change_type:
                # We need to disable type checking here since gitpython expects a value
                # of type `Lit_change_type` in iter_change_type() but returns plain
                # strings in `diff.change_type`.
                for c in diff.iter_change_type(ct):  # type: ignore[arg-type]
                    difftext.extend(_process_diff(ct, c, diff_func))

        return "\n".join(difftext), changed

    def stage_files(self, files: Sequence[str]):
        """Add provided list of files to index."""
        self._repo.index.add(files)

    def commit(self, commit_message: str, amend=False):
        author = self._author

        if self._trace:
            click.echo(f' > Using "{author.name} <{author.email}>" as commit author')

        if amend:
            # We need to call out to `git commit` for amending
            self._repo.git.execute(  # type: ignore[call-overload]
                [
                    "git",
                    "commit",
                    "--amend",
                    "--no-edit",
                    "--reset-author",
                    "-m",
                    commit_message,
                ],
                env=self._author_env,
            )
        else:
            self._repo.index.commit(commit_message, author=author, committer=author)

    def push(
        self, remote: Optional[str] = None, version: Optional[str] = None
    ) -> Iterable[PushInfo]:
        if not remote:
            remote = "origin"
        if not version:
            version = self._default_version()
        return self._repo.remote(remote).push(version)

    def reset(self, working_tree: bool = False):
        self._repo.head.reset(working_tree=working_tree)



from __future__ import annotations

import shutil
import tempfile

from pathlib import Path

import click

from commodore.config import Config

from .parameters import (
    ClassNotFound,
    InventoryFactory,
    InventoryFacts,
    InventoryParameters,
)


def _cleanup_work_dir(cfg: Config, work_dir: Path):
    if not cfg.debug:
        # Clean up work dir if we're not in debug mode
        shutil.rmtree(work_dir)


def extract_packages(
    cfg: Config, invfacts: InventoryFacts
) -> dict[str, dict[str, str]]:
    return _get_inventory(cfg, invfacts).parameters("packages")


def extract_components(
    cfg: Config, invfacts: InventoryFacts
) -> dict[str, dict[str, str]]:
    return _get_inventory(cfg, invfacts).parameters("components")


def extract_parameters(
    cfg: Config, invfacts: InventoryFacts
) -> dict[str, dict[str, str]]:
    return _get_inventory(cfg, invfacts).parameters()


def _get_inventory(cfg: Config, invfacts: InventoryFacts) -> InventoryParameters:
    if cfg.debug:
        click.echo(
            f"Called with: global_config={invfacts.global_config} "
            + f"tenant_config={invfacts.tenant_config} "
            + f"extra_classes={invfacts.extra_classes} "
            + f"allow_missing_classes={invfacts.allow_missing_classes}."
        )

    global_dir = Path(invfacts.global_config).resolve().absolute()
    tenant_dir = None
    if invfacts.tenant_config:
        tenant_dir = Path(invfacts.tenant_config).resolve().absolute()

    work_dir = Path(tempfile.mkdtemp(prefix="commodore-reclass-")).resolve()

    if global_dir.is_dir() and (not tenant_dir or tenant_dir.is_dir()):
        invfactory = InventoryFactory.from_repo_dirs(
            work_dir, global_dir, tenant_dir, invfacts
        )
    else:
        _cleanup_work_dir(cfg, work_dir)
        raise NotImplementedError("Cloning global or tenant repo not yet implemented")

    try:
        inv = invfactory.reclass(invfacts)
    except ClassNotFound as e:
        _cleanup_work_dir(cfg, work_dir)
        raise ValueError(
            "Unable to render inventory with `--no-allow-missing-classes`. "
            + f"Class '{e.name}' not found. "
            + "Verify the provided values or allow missing classes."
        ) from e

    _cleanup_work_dir(cfg, work_dir)

    return inv



from __future__ import annotations

import json
import time
import textwrap

from enum import Enum
from pathlib import Path as P
from typing import Any, Iterable, Optional

import click
import jwt
import requests

from url_normalize import url_normalize

from commodore.component import Component, component_parameters_key
from .gitrepo import GitRepo
from .inventory import Inventory
from .multi_dependency import MultiDependency, dependency_key
from .package import Package
from . import tokencache


class Migration(Enum):
    KAP_029_030 = "kapitan-0.29-to-0.30"


# pylint: disable=too-many-instance-attributes,too-many-public-methods
class Config:
    _inventory: Inventory
    _components: dict[str, Component]
    _config_repos: dict[str, GitRepo]
    _component_aliases: dict[str, str]
    _packages: dict[str, Package]
    _dependency_repos: dict[str, MultiDependency]
    _deprecation_notices: list[str]
    _migration: Optional[Migration]
    _dynamic_facts: dict[str, Any]
    _github_token: Optional[str]

    oidc_client: Optional[str]
    oidc_discovery_url: Optional[str]

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        work_dir: P,
        api_url=None,
        api_token=None,
        verbose=0,
        username=None,
        usermail=None,
    ):
        self._work_dir = work_dir.resolve()
        self.api_url = api_url
        self.api_token = api_token
        self.oidc_client = None
        self.oidc_discovery_url = None
        self._components = {}
        self._config_repos = {}
        self._component_aliases = {}
        self._packages = {}
        self._dependency_repos = {}
        self._verbose = verbose
        self.username = username
        self.usermail = usermail
        self.local = False
        self.push = None
        self.interactive = None
        self.force = False
        self._fetch_dependencies = True
        self._inventory = Inventory(work_dir=self.work_dir)
        self._deprecation_notices = []
        self._global_repo_revision_override = None
        self._tenant_repo_revision_override = None
        self._migration = None
        self._dynamic_facts = {}
        self._github_token = None

    @property
    def verbose(self):
        return self._verbose

    @property
    def debug(self):
        return self._verbose > 0

    @property
    def trace(self):
        return self._verbose >= 3

    @property
    def config_file(self):
        return self._inventory.global_config_dir / "commodore.yml"

    @property
    def jsonnet_file(self) -> P:
        return self._work_dir / "jsonnetfile.json"

    @property
    def work_dir(self) -> P:
        return self._work_dir

    @work_dir.setter
    def work_dir(self, d: P):
        self._work_dir = d
        self.inventory.work_dir = d

    @property
    def vendor_dir(self) -> P:
        return self.work_dir / "vendor"

    @property
    def catalog_dir(self) -> P:
        return self.work_dir / "catalog"

    @property
    def refs_dir(self) -> P:
        return self.catalog_dir / "refs"

    @property
    def fetch_dependencies(self) -> bool:
        return self._fetch_dependencies

    @fetch_dependencies.setter
    def fetch_dependencies(self, value: bool):
        if not self.local:
            if not value:
                click.secho(
                    "[WARN] --no-fetch-dependencies doesn't take effect "
                    + "unless --local is specified",
                    fg="yellow",
                )
            value = True
        self._fetch_dependencies = value

    @property
    def api_token(self):
        if self._api_token is None and self.api_url:
            tokens = tokencache.get(self.api_url)
            token = tokens.get("id_token")
            if token is not None:
                # We don't verify the signature, we just want to know if the token is expired
                # lieutenant will decide if it's valid
                try:
                    t = jwt.decode(
                        token, algorithms=["RS256"], options={"verify_signature": False}
                    )
                    if "exp" in t and t["exp"] < time.time() + 10:
                        return None
                except jwt.exceptions.InvalidTokenError:
                    return None
                self._api_token = token
        return self._api_token

    @api_token.setter
    def api_token(self, api_token):
        if api_token is not None:
            try:
                p = P(api_token)
                if p.is_file():
                    with open(p, encoding="utf-8") as apitoken:
                        api_token = apitoken.read()
            except OSError as e:
                # File name too long, assume token is not configured as file
                if "File name too long" in e.strerror:
                    pass
                else:
                    raise
            self._api_token = api_token.strip()
        else:
            self._api_token = None

    @property
    def global_repo_revision_override(self):
        return self._global_repo_revision_override

    @global_repo_revision_override.setter
    def global_repo_revision_override(self, rev):
        self._global_repo_revision_override = rev

    @property
    def tenant_repo_revision_override(self):
        return self._tenant_repo_revision_override

    @tenant_repo_revision_override.setter
    def tenant_repo_revision_override(self, rev):
        self._tenant_repo_revision_override = rev

    @property
    def migration(self):
        return self._migration

    @migration.setter
    def migration(self, migration):
        if migration and migration != "":
            self._migration = Migration(migration)

    @property
    def dynamic_facts(self) -> dict[str, Any]:
        """Returns fallback dynamic facts provided on the command line."""
        return self._dynamic_facts

    @dynamic_facts.setter
    def dynamic_facts(self, facts: dict[str, Any]):
        self._dynamic_facts = facts

    @property
    def github_token(self) -> Optional[str]:
        return self._github_token

    @github_token.setter
    def github_token(self, github_token: str):
        self._github_token = github_token

    @property
    def inventory(self):
        return self._inventory

    def update_verbosity(self, verbose):
        self._verbose += verbose

    def get_components(self):
        return self._components

    def register_component(self, component: Component):
        self._components[component.name] = component

    def get_component_repo(self, component_name):
        return self._components[component_name].repo

    def get_configs(self):
        return self._config_repos

    def register_config(self, level, repo: GitRepo):
        self._config_repos[level] = repo

    def get_packages(self) -> dict[str, Package]:
        return self._packages

    def register_package(self, pkg_name: str, pkg: Package):
        self._packages[pkg_name] = pkg

    def register_dependency_repo(self, repo_url: str) -> MultiDependency:
        """Register dependency repository, if it isn't registered yet.

        Returns the `MultiDependency` object for the repo."""
        depkey = dependency_key(repo_url)
        if depkey not in self._dependency_repos:
            self._dependency_repos[depkey] = MultiDependency(
                repo_url, self.inventory.dependencies_dir
            )

        dep = self._dependency_repos[depkey]
        # Prefer ssh fetch URLs for existing dependencies
        if repo_url.startswith("ssh://"):
            dep.url = repo_url
        return dep

    def get_component_aliases(self):
        return self._component_aliases

    def register_component_aliases(self, aliases: dict[str, str]):
        self._component_aliases = aliases

    def verify_component_aliases(self, cluster_parameters: dict):
        for alias, cn in self._component_aliases.items():
            if alias != cn and not _component_is_aliasable(cluster_parameters, cn):
                raise click.ClickException(
                    f"Component {cn} with alias {alias} does not support instantiation."
                )

    def register_deprecation_notice(self, notice: str):
        self._deprecation_notices.append(notice)

    def print_deprecation_notices(self):
        tw = textwrap.TextWrapper(
            width=100,
            # Next two options ensure we don't break URLs
            break_long_words=False,
            break_on_hyphens=False,
            initial_indent=" > ",
            subsequent_indent="   ",
        )
        if len(self._deprecation_notices) > 0:
            click.secho("\nCommodore notices:", bold=True)
            for notice in self._deprecation_notices:
                notice = tw.fill(notice)
                click.secho(notice)

    def register_component_deprecations(self, cluster_parameters):
        for cname in self._component_aliases.values():
            ckey = component_parameters_key(cname)
            cmeta = cluster_parameters[ckey].get("_metadata", {})

            if cmeta.get("deprecated", False):
                msg = f"Component {cname} is deprecated."
                if "replaced_by" in cmeta:
                    msg += f" Use component {cmeta['replaced_by']} instead."
                if "deprecation_notice" in cmeta:
                    msg += f" {cmeta['deprecation_notice']}"
                self.register_deprecation_notice(msg)

    def discover_oidc_config(self) -> None:
        """Check the configured Lieutenant API URL for OIDC client details, if no OIDC
        client details are given on the command line.

        Update the provided config object in place if the API provides OIDC client
        details."""
        if (
            self.oidc_client is None
            and self.oidc_discovery_url is None
            and self.api_url is not None
        ):
            try:
                r = requests.get(url_normalize(self.api_url))
                api_cfg = json.loads(r.text)
                if "oidc" in api_cfg:
                    self.oidc_client = api_cfg["oidc"].get("clientId")
                    self.oidc_discovery_url = api_cfg["oidc"].get("discoveryUrl")
            except (requests.RequestException, json.JSONDecodeError) as e:
                # We do this on a best effort basis
                click.echo(f" > Unable to auto-discover OIDC config: {e}")


def _component_is_aliasable(cluster_parameters: dict, component_name: str):
    ckey = component_parameters_key(component_name)
    cmeta = cluster_parameters[ckey].get("_metadata", {})
    return cmeta.get("multi_instance", False)


def set_fact_value(facts: dict[str, Any], raw_key: str, value: Any) -> None:
    """Set value for nested fact at `raw_key` (expected form `path.to.key`) to `value`.

    If a segment of the nested key is present in `facts` and not a dictionary or when
    the raw key contains an empty segment, the function won't set a value and will
    instead print a diagnostic message.
    """
    key_parts = raw_key.split(".")

    if any(kp == "" for kp in key_parts):
        # Bail out early if the raw key is malformed (any empty segment)
        click.secho(f"Malformed nested key '{raw_key}' skipping...", fg="yellow")
        return

    prefix_key = ""
    target_dict = facts
    for k in key_parts[:-1]:
        prefix_key = f"{prefix_key}{k}."
        if k in target_dict and not isinstance(target_dict[k], dict):
            click.secho(
                "Trying to insert subkey into non-dictionary "
                + f"dynamic fact '{prefix_key[:-1]}', skipping...",
                fg="yellow",
            )
            return

        target_dict = target_dict.setdefault(k, {})

    key = key_parts[-1]

    if key in target_dict:
        click.secho(
            f"Overwriting dynamic fact '{raw_key}={target_dict[key]}' with '{value}'",
            fg="yellow",
        )
    target_dict[key] = value


def parse_dynamic_fact_value(raw_value: str) -> Any:
    """Parse raw dynamic fact value.

    Tries to parse the value as JSON if it starts with the literal `json:`.

    Returns the parsed value or `None` if trying to parse a JSON value results in a
    decode error."""
    if raw_value.startswith("json:"):
        json_val = raw_value.replace("json:", "", 1)
        # Parse value as JSON if it starts with `json:`, skip value completely
        # on parse errors.
        try:
            v = json.loads(json_val)
        except json.JSONDecodeError as e:
            click.secho(
                f"Expected value '{json_val}' to be parsable JSON, "
                + f"but parsing failed with '{e}', skipping",
                fg="yellow",
            )
            return None
    else:
        v = raw_value
    return v


def parse_dynamic_facts_from_cli(raw_facts: Iterable[str]) -> dict[str, Any]:
    """Parse dynamic facts dictionary from strings provided on command line.

    The function expects each raw fact (string) to be of the form `key=value`, where key
    can contain dots to specify nested keys (e.g. `path.to.key`), and value is parsed as
    JSON when it starts with the literal `json:`.

    Inputs are processed in order, and subsequent inputs setting the same key will
    overwrite any existing values. Facts for nested keys will be skipped, if a parent
    key already exists *and* isn't a dictionary.

    Facts with values marked as JSON which result in a decode error will be skipped.

    Returns a dict with the parsed dynamic facts structure.
    """
    facts: dict[str, Any] = {}

    for f in raw_facts:
        if "=" not in f:
            click.secho(
                f"Ignoring dynamic fact {f} which is not in format key=value",
                fg="yellow",
            )
            continue
        raw_key, raw_value = f.split("=", maxsplit=1)
        if not raw_key:
            click.secho(
                f"Ignoring malformed dynamic fact '{f}' with no key.", fg="yellow"
            )
            continue
        if not raw_value:
            click.secho(
                f"Ignoring malformed dynamic fact '{f}' with no value. "
                + "Please specify empty string value as 'json:\"\"'",
                fg="yellow",
            )
            continue

        # Parse value first, so we never add empty dicts for a nested key whose JSON
        # value turns out to be malformed.
        value = parse_dynamic_fact_value(raw_value)
        if value is None:
            # skip when we failed to parse a value that identified itself as JSON
            continue

        set_fact_value(facts, raw_key, value)

    return facts



from __future__ import annotations

import json
import sys
from collections.abc import Iterable
from pathlib import Path
from typing import Optional

import click
import yaml

from dotenv import load_dotenv, find_dotenv
from commodore import __git_version__, __version__
from .catalog import catalog_list
from .config import Config, Migration, parse_dynamic_facts_from_cli
from .helpers import clean_working_tree
from .compile import compile as _compile
from .component.template import ComponentTemplater
from .component.compile import compile_component
from .inventory.render import extract_components, extract_packages, extract_parameters
from .inventory.parameters import InventoryFacts
from .inventory.lint import LINTERS
from .login import login, fetch_token
from .package.compile import compile_package
from .package.template import PackageTemplater
from .package.sync import sync_packages

pass_config = click.make_pass_decorator(Config)

verbosity = click.option(
    "-v",
    "--verbose",
    count=True,
    help="Control verbosity. Can be repeated for more verbose output.",
)


def _version():
    if f"v{__version__}" != __git_version__:
        return f"{__version__} (Git version: {__git_version__})"
    return __version__


CONTEXT_SETTINGS = {"help_option_names": ["-h", "--help"]}


@click.group(context_settings=CONTEXT_SETTINGS)
@click.version_option(_version(), prog_name="commodore")
@verbosity
@click.option(
    "-d",
    "--working-dir",
    default="./",
    show_default=True,
    type=click.Path(file_okay=False, dir_okay=True),
    envvar="COMMODORE_WORKING_DIR",
    help=(
        "The directory in which Commodore will fetch dependencies, "
        "inventory and catalog, and store intermediate outputs"
    ),
)
@click.pass_context
def commodore(ctx, working_dir, verbose):
    ctx.obj = Config(Path(working_dir), verbose=verbose)


@commodore.group(short_help="Interact with a cluster catalog.")
@verbosity
@pass_config
def catalog(config: Config, verbose):
    config.update_verbosity(verbose)


@catalog.command(short_help="Delete generated files.")
@verbosity
@pass_config
def clean(config: Config, verbose):
    config.update_verbosity(verbose)
    clean_working_tree(config)


api_url_option = click.option(
    "--api-url", envvar="COMMODORE_API_URL", help="Lieutenant API URL.", metavar="URL"
)
oidc_discovery_url_option = click.option(
    "--oidc-discovery-url",
    envvar="COMMODORE_OIDC_DISCOVERY_URL",
    help="The discovery URL of the IdP.",
    metavar="URL",
)
oidc_client_option = click.option(
    "--oidc-client",
    envvar="COMMODORE_OIDC_CLIENT",
    help="The OIDC client name.",
    metavar="TEXT",
)


@catalog.command(name="compile", short_help="Compile the catalog.")
@click.argument("cluster")
@api_url_option
@click.option(
    "--api-token",
    envvar="COMMODORE_API_TOKEN",
    help="Lieutenant API token.",
    metavar="TOKEN",
)
@oidc_discovery_url_option
@oidc_client_option
@click.option(
    "--local",
    is_flag=True,
    default=False,
    help=(
        "Run in local mode, local mode does not try to connect to "
        + "the Lieutenant API or fetch/push Git repositories."
    ),
)
@click.option(
    "--push", is_flag=True, default=False, help="Push catalog to remote repository."
)
@click.option(
    "-i",
    "--interactive",
    is_flag=True,
    default=False,
    help="Prompt confirmation to push to remote repository.",
)
@click.option(
    "--git-author-name",
    envvar="GIT_AUTHOR_NAME",
    metavar="USERNAME",
    help="Name of catalog commit author",
)
@click.option(
    "--git-author-email",
    envvar="GIT_AUTHOR_EMAIL",
    metavar="EMAIL",
    help="E-mail address of catalog commit author",
)
@click.option(
    "-g",
    "--global-repo-revision-override",
    envvar="GLOBAL_REPO_REVISION_OVERRIDE",
    metavar="REV",
    help=(
        "Git revision (tree-ish) to checkout for the global config repo "
        + "(overrides configuration in Lieutenant tenant & cluster)"
    ),
)
@click.option(
    "-t",
    "--tenant-repo-revision-override",
    envvar="TENANT_REPO_REVISION_OVERRIDE",
    metavar="REV",
    help=(
        "Git revision (tree-ish) to checkout for the tenant config repo "
        + "(overrides configuration in Lieutenant cluster)"
    ),
)
@click.option(
    " / -F",
    "--fetch-dependencies/--no-fetch-dependencies",
    default=True,
    help=(
        "Whether to fetch Jsonnet and Kapitan dependencies in local mode. "
        + "By default dependencies are fetched."
    ),
)
@click.option(
    "-m",
    "--migration",
    help=(
        "Specify a migration that you expect to happen for the cluster catalog. "
        + "Currently Commodore only knows the Kapitan 0.29 to 0.30 migration. "
        + "When the Kapitan 0.29 to 0.30 migration is selected, Commodore will suppress "
        + "noise (changing managed-by labels, and reordered objects) caused by the "
        + "migration in the diff output."
    ),
    type=click.Choice([m.value for m in Migration], case_sensitive=False),
)
@click.option(
    "-d",
    "--dynamic-fact",
    type=str,
    metavar="KEY=VALUE",
    multiple=True,
    help=(
        "Fallback dynamic facts to use when compiling a cluster which hasn't "
        + "reported its dynamic facts yet. Commodore will never use values provided "
        + "through this parameter if the cluster response from the API has a dynamic "
        + "facts field. Can be repeated. Commodore expects each fact to be specified "
        + "as key=value. Nested keys can be provided as `path.to.key`. Commodore will "
        + "parse values as JSON if they're prefixed by `json:`. If the same key is "
        + "provided multiple times, the last occurrence overrides the previous values. "
        + "When providing a value for a key as JSON, previously specified subkeys of "
        + "that key will be overwritten. Nested keys are ignored if any non-leaf level "
        + "of the requested key already contains a non-dictionary value. If a value "
        + "prefixed with `json:` isn't valid JSON, it will be skipped."
    ),
)
@verbosity
@pass_config
# pylint: disable=too-many-arguments
# pylint: disable=too-many-locals
def compile_catalog(
    config: Config,
    cluster,
    api_url,
    api_token,
    oidc_client,
    oidc_discovery_url,
    local,
    push,
    interactive,
    verbose,
    git_author_name,
    git_author_email,
    global_repo_revision_override,
    tenant_repo_revision_override,
    fetch_dependencies,
    migration,
    dynamic_fact: str,
):
    config.update_verbosity(verbose)
    config.api_url = api_url
    config.api_token = api_token
    config.local = local
    config.push = push
    config.interactive = interactive
    config.username = git_author_name
    config.usermail = git_author_email
    config.global_repo_revision_override = global_repo_revision_override
    config.tenant_repo_revision_override = tenant_repo_revision_override
    config.migration = migration
    config.oidc_client = oidc_client
    config.oidc_discovery_url = oidc_discovery_url
    config.fetch_dependencies = fetch_dependencies
    config.dynamic_facts = parse_dynamic_facts_from_cli(dynamic_fact)

    if config.push and (
        config.global_repo_revision_override or config.tenant_repo_revision_override
    ):
        raise click.ClickException(
            "Cannot push changes when local global or tenant repo override is specified"
        )

    if config.api_token is None and not local:
        try:
            login(config)
        except click.ClickException:
            pass

    _compile(config, cluster)


@catalog.command(name="list", short_help="List available catalog cluster IDs")
@api_url_option
@click.option(
    "--api-token",
    envvar="COMMODORE_API_TOKEN",
    help="Lieutenant API token.",
    metavar="TOKEN",
)
@oidc_client_option
@oidc_discovery_url_option
@verbosity
@pass_config
# pylint: disable=too-many-arguments
def clusters_list_command(
    config: Config, api_url, api_token, oidc_client, oidc_discovery_url, verbose
):
    config.update_verbosity(verbose)
    config.api_url = api_url
    config.api_token = api_token
    config.oidc_client = oidc_client
    config.oidc_discovery_url = oidc_discovery_url

    if config.api_token is None:
        try:
            login(config)
        except click.ClickException:
            pass

    catalog_list(config)


@commodore.group(short_help="Interact with components.")
@verbosity
@pass_config
def component(config: Config, verbose):
    config.update_verbosity(verbose)


@component.command(name="new", short_help="Bootstrap a new component.")
@click.argument("slug")
@click.option(
    "--name",
    help="The component's name as it will be written in the documentation. Defaults to the slug.",
)
@click.option(
    "--lib/--no-lib",
    default=False,
    show_default=True,
    help="Add a component library template.",
)
@click.option(
    "--pp/--no-pp",
    default=False,
    show_default=True,
    help="Add a component postprocessing template.",
)
@click.option(
    "--owner",
    default="projectsyn",
    show_default=True,
    help="The GitHub user or project name where the component will be hosted.",
)
@click.option(
    "--copyright",
    "copyright_holder",
    default="VSHN AG <info@vshn.ch>",
    show_default=True,
    help="The copyright holder added to the license file.",
)
@click.option(
    "--golden-tests/--no-golden-tests",
    default=True,
    show_default=True,
    help="Add golden tests to the component.",
)
@click.option(
    "--matrix-tests/--no-matrix-tests",
    default=True,
    show_default=True,
    help="Enable test matrix for compile/golden tests.",
)
@verbosity
@pass_config
# pylint: disable=too-many-arguments
def component_new(
    config: Config,
    slug,
    name,
    lib,
    pp,
    owner,
    copyright_holder,
    golden_tests,
    matrix_tests,
    verbose,
):
    config.update_verbosity(verbose)
    f = ComponentTemplater(config, slug, name=name)
    f.library = lib
    f.post_process = pp
    f.github_owner = owner
    f.copyright_holder = copyright_holder
    f.golden_tests = golden_tests
    f.matrix_tests = matrix_tests
    f.create()


@component.command(name="delete", short_help="Remove component from inventory.")
@click.argument("slug")
@click.option(
    "--force/--no-force",
    default=False,
    show_default=True,
    help="Don't prompt for user confirmation when deleting.",
)
@verbosity
@pass_config
# pylint: disable=too-many-arguments
def component_delete(config: Config, slug, force, verbose):
    config.update_verbosity(verbose)
    config.force = force
    f = ComponentTemplater(config, slug)
    f.delete()


@component.command(name="compile", short_help="Compile a single component.")
@click.argument("path", type=click.Path(exists=True, file_okay=False, dir_okay=True))
@click.option(
    "-f",
    "--values",
    multiple=True,
    type=click.Path(exists=True, file_okay=True, dir_okay=False),
    help="Specify inventory class in a YAML file (can specify multiple).",
)
@click.option(
    "-a",
    "--alias",
    metavar="ALIAS",
    help="Provide component alias to use when compiling component.",
)
@click.option(
    "-J",
    "--search-paths",
    multiple=True,
    type=click.Path(exists=True, file_okay=False, dir_okay=True),
    help="Specify additional search paths.",
)
@click.option(
    "-o",
    "--output",
    default="./",
    show_default=True,
    type=click.Path(file_okay=False, dir_okay=True),
    help="Specify output path for compiled component.",
)
@click.option(
    "-r",
    "--repo-directory",
    default="",
    show_default=False,
    type=click.Path(file_okay=False, dir_okay=True),
    help="Path to the root of the Git repo containing the component. "
    + "The command assumes that the component is in the repo root, "
    + "if this option is not provided.",
)
@click.option(
    "-n",
    "--name",
    default="",
    show_default=False,
    help="Component name to use for Commodore. "
    + "If not provided, the name is inferred from the Git repository name.",
)
@verbosity
@pass_config
# pylint: disable=too-many-arguments
def component_compile(
    config: Config,
    path: str,
    values: Iterable[str],
    alias: Optional[str],
    search_paths: Iterable[str],
    output: str,
    repo_directory: str,
    name: str,
    verbose: int,
):
    config.update_verbosity(verbose)
    compile_component(
        config, path, alias, values, search_paths, output, repo_directory, name
    )


@commodore.group(short_help="Interact with a Commodore config package")
@verbosity
@pass_config
def package(config: Config, verbose: int):
    config.update_verbosity(verbose)


@package.command(name="new", short_help="Create a new config package from a template")
@click.argument("slug")
@click.option(
    "--name",
    help="The package's name as it will be written in the documentation. Defaults to the slug.",
)
@click.option(
    "--owner",
    default="projectsyn",
    show_default=True,
    help="The GitHub user or project name where the package will be hosted.",
)
@click.option(
    "--copyright",
    "copyright_holder",
    default="VSHN AG <info@vshn.ch>",
    show_default=True,
    help="The copyright holder added to the license file.",
)
@click.option(
    "--golden-tests/--no-golden-tests",
    default=True,
    show_default=True,
    help="Add golden tests to the package.",
)
@click.option(
    "--template-url",
    default="https://github.com/projectsyn/commodore-config-package-template.git",
    show_default=True,
    help="The URL of the package cookiecutter template.",
)
@click.option(
    "--template-version",
    default="main",
    show_default=True,
    help="The package template version (Git tree-ish) to use.",
)
@click.option(
    "--output-dir",
    default="",
    show_default=True,
    type=click.Path(file_okay=False, dir_okay=True),
    help="The directory in which to place the new package.",
)
@click.option(
    "--additional-test-case",
    "-t",
    metavar="CASE",
    default=[],
    show_default=True,
    multiple=True,
    help="Additional test cases to generate in the new package. Can be repeated. "
    + "Test case `defaults` will always be generated."
    + "Commodore will deduplicate test cases by name.",
)
@verbosity
@pass_config
# pylint: disable=too-many-arguments
def package_new(
    config: Config,
    verbose: int,
    slug: str,
    name: Optional[str],
    owner: str,
    copyright_holder: str,
    golden_tests: bool,
    template_url: str,
    template_version: str,
    output_dir: str,
    additional_test_case: Iterable[str],
):
    """Create new config package repo from template.

    The command line options allow the caller to customize the remote repository
    location (only supports GitHub at the moment), the package's display name, whether
    to configure golden tests, and the licensing details.
    """
    config.update_verbosity(verbose)
    t = PackageTemplater(config, slug, name=name, output_dir=output_dir)
    t.github_owner = owner
    t.copyright_holder = copyright_holder
    t.golden_tests = golden_tests
    t.template_url = template_url
    t.template_version = template_version
    t.test_cases = ["defaults"] + list(additional_test_case)
    t.create()


@package.command(
    name="update", short_help="Update an existing config package from a template"
)
@click.argument(
    "package_path", type=click.Path(exists=True, dir_okay=True, file_okay=False)
)
@click.option(
    "--copyright",
    "copyright_holder",
    show_default=True,
    help="The copyright holder added to the license file.",
)
@click.option(
    "--golden-tests/--no-golden-tests",
    default=None,
    show_default=True,
    help="Add golden tests to the package.",
)
@click.option(
    "--update-copyright-year/--no-update-copyright-year",
    default=False,
    show_default=True,
    help="Update year in copyright notice.",
)
@click.option(
    "--additional-test-case",
    "-t",
    metavar="CASE",
    default=[],
    show_default=True,
    multiple=True,
    help="Additional test cases to generate in the new package. Can be repeated. "
    + "Already existing test cases will always be kept. "
    + "Commodore will deduplicate test cases by name.",
)
@click.option(
    "--remove-test-case",
    metavar="CASE",
    default=[],
    show_default=True,
    multiple=True,
    help="Test cases to remove from the package. Can be repeated.",
)
@verbosity
@pass_config
# pylint: disable=too-many-arguments
def package_update(
    config: Config,
    verbose: int,
    package_path: str,
    copyright_holder: Optional[str],
    golden_tests: Optional[bool],
    update_copyright_year: bool,
    additional_test_case: Iterable[str],
    remove_test_case: Iterable[str],
):
    config.update_verbosity(verbose)
    t = PackageTemplater.from_existing(config, Path(package_path))

    # Add provided values
    if copyright_holder:
        t.copyright_holder = copyright_holder
    if golden_tests is not None:
        t.golden_tests = golden_tests
    if update_copyright_year:
        t.copyright_year = None
    test_cases = t.test_cases
    test_cases.extend(additional_test_case)
    t.test_cases = [tc for tc in test_cases if tc not in remove_test_case]

    t.update()


@package.command(name="compile", short_help="Compile a config package standalone")
@click.argument("path", type=click.Path(exists=True, file_okay=False, dir_okay=True))
@click.argument(
    "test_class", type=click.Path(exists=False, file_okay=True, dir_okay=False)
)
@click.option(
    "-f",
    "--values",
    multiple=True,
    type=click.Path(exists=True, file_okay=True, dir_okay=False),
    help="Specify additional inventory class in a YAML file (can specify multiple).",
)
@click.option(
    "--local",
    is_flag=True,
    default=False,
    help=(
        "Run in local mode, local mode reuses the contents of the working directory. "
        + "Local mode won't fetch missing components."
    ),
)
@click.option(
    " / -F",
    "--fetch-dependencies/--no-fetch-dependencies",
    default=True,
    help="Whether to fetch Jsonnet and Kapitan dependencies in local mode. "
    + "By default dependencies are fetched.",
)
@click.option(
    "-k / ",
    "--keep-dir/--no-keep-dir",
    default=False,
    show_default=True,
    help="Whether to keep the compilation temp directory after the compilation is done.",
)
@click.option(
    "--tmp-dir",
    default="",
    metavar="PATH",
    type=click.Path(exists=False, file_okay=False, dir_okay=True),
    help="Temp directory to use for compilation. Implies `--keep-dir`",
)
@verbosity
@pass_config
# pylint: disable=too-many-arguments
def package_compile(
    config: Config,
    verbose: int,
    path: str,
    test_class: str,
    values: Iterable[str],
    local: bool,
    fetch_dependencies: bool,
    keep_dir: bool,
    tmp_dir: str,
):
    config.update_verbosity(verbose)
    config.local = local
    config.fetch_dependencies = fetch_dependencies
    compile_package(config, path, test_class, values, tmp_dir, keep_dir)


@package.command("sync", short_help="Synchronize packages to template")
@verbosity
@pass_config
@click.argument(
    "package_list", type=click.Path(file_okay=True, dir_okay=False, exists=True)
)
@click.option(
    "--github-token",
    help="GitHub API token",
    envvar="COMMODORE_GITHUB_TOKEN",
    default="",
)
@click.option(
    "--dry-run", is_flag=True, help="Don't create or update PRs", default=False
)
@click.option(
    "--pr-branch",
    "-b",
    metavar="BRANCH",
    default="template-sync",
    type=str,
    help="Branch name to use for updates from template",
)
@click.option(
    "--pr-label",
    "-l",
    metavar="LABEL",
    default=[],
    multiple=True,
    help="Labels to set on the PR. Can be repeated",
)
def package_sync(
    config: Config,
    verbose: int,
    package_list: str,
    github_token: str,
    dry_run: bool,
    pr_branch: str,
    pr_label: Iterable[str],
):
    """This command processes all packages listed in the provided `PACKAGE_LIST` YAML file.

    Currently, the command only supports updating packages hosted on GitHub. The command
    expects that the YAML file contains a single document with a list of GitHub
    repositories in form `organization/repository-name`.

    The command clones each package and runs `package update` on the local copy. If
    there are any changes, the command creates a PR for the changes. For each package,
    the command parses the package's `.cruft.json` to determine the template repository
    and template version for the package. The command bases each PR on the default
    branch of the corresponding package repository as reported by the GitHub API.

    The command requires a GitHub Access token with the 'public_repo' permission, which
    is required to create PRs on public repositories. If you want to manage private
    repos, the access token may require additional permissions.
    """
    config.update_verbosity(verbose)
    config.github_token = github_token

    sync_packages(config, Path(package_list), dry_run, pr_branch, pr_label)


@commodore.group(short_help="Interact with a Commodore inventory")
@verbosity
@pass_config
def inventory(config: Config, verbose):
    config.update_verbosity(verbose)


inventory_output_format = click.option(
    "-o",
    "--output-format",
    help="Output format",
    type=click.Choice(["json", "yaml"]),
    default="yaml",
)
inventory_values = click.option(
    "-f",
    "--values",
    help=(
        "Extra values file to use when rendering inventory. "
        + "Used as additional reclass class. "
        + "Use a values file to specify any cluster facts. "
        + "Can be repeated."
    ),
    multiple=True,
    type=click.Path(exists=True, file_okay=True, dir_okay=False),
)
inventory_allow_missing_classes = click.option(
    " / -A",
    "--allow-missing-classes/--no-allow-missing-classes",
    default=True,
    help="Whether to allow missing classes when rendering the inventory. Defaults to true.",
)


@inventory.command(
    name="show",
    short_help="Returns the rendered inventory",
)
@inventory_output_format
@inventory_values
@inventory_allow_missing_classes
@click.argument("global-config")
@click.argument("tenant-config", required=False)
@verbosity
@pass_config
# pylint: disable=too-many-arguments
def inventory_show(
    config: Config,
    verbose,
    global_config: str,
    tenant_config: Optional[str],
    output_format: str,
    values: Iterable[str],
    allow_missing_classes: bool,
):
    config.update_verbosity(verbose)
    extra_values = [Path(v) for v in values]
    try:
        inv = extract_parameters(
            config,
            InventoryFacts(
                global_config,
                tenant_config,
                extra_values,
                allow_missing_classes,
                ignore_class_notfound_warning=False,
            ),
        )
    except ValueError as e:
        raise click.ClickException(f"While rendering inventory: {e}") from e

    if output_format == "json":
        click.echo(json.dumps(inv))
    else:
        click.echo(yaml.safe_dump(inv))


@inventory.command(
    name="components",
    short_help="Extract component URLs and versions from the inventory",
)
@inventory_output_format
@inventory_values
@inventory_allow_missing_classes
@click.argument("global-config")
@click.argument("tenant-config", required=False)
@verbosity
@pass_config
# pylint: disable=too-many-arguments
def component_versions(
    config: Config,
    verbose,
    global_config: str,
    tenant_config: Optional[str],
    output_format: str,
    values: Iterable[str],
    allow_missing_classes: bool,
):
    config.update_verbosity(verbose)
    extra_values = [Path(v) for v in values]
    try:
        components = extract_components(
            config,
            InventoryFacts(
                global_config,
                tenant_config,
                extra_values,
                allow_missing_classes,
                ignore_class_notfound_warning=False,
            ),
        )
    except ValueError as e:
        raise click.ClickException(f"While extracting components: {e}") from e

    if output_format == "json":
        click.echo(json.dumps(components))
    else:
        click.echo(yaml.safe_dump(components))


@inventory.command(
    name="packages",
    short_help="Extract package URLs and versions from the inventory",
)
@inventory_output_format
@inventory_values
@inventory_allow_missing_classes
@click.argument("global-config")
@click.argument("tenant-config", required=False)
@verbosity
@pass_config
# pylint: disable=too-many-arguments
def package_versions(
    config: Config,
    verbose,
    global_config: str,
    tenant_config: Optional[str],
    output_format: str,
    values: Iterable[str],
    allow_missing_classes: bool,
):
    config.update_verbosity(verbose)
    extra_values = [Path(v) for v in values]
    try:
        pkgs = extract_packages(
            config,
            InventoryFacts(
                global_config,
                tenant_config,
                extra_values,
                allow_missing_classes,
                ignore_class_notfound_warning=False,
            ),
        )
    except ValueError as e:
        raise click.ClickException(f"While extracting packages: {e}") from e

    if output_format == "json":
        click.echo(json.dumps(pkgs))
    else:
        click.echo(yaml.safe_dump(pkgs))


@inventory.command(
    name="lint",
    short_help="Lint YAML files for Commodore inventory structures in the provided paths",
    no_args_is_help=True,
)
@click.option(
    "-l",
    "--linter",
    help="Which linters to enable. Can be repeated.",
    type=click.Choice(
        list(LINTERS.keys()),
        case_sensitive=False,
    ),
    multiple=True,
    default=tuple(LINTERS.keys()),
)
@click.argument(
    "target", type=click.Path(file_okay=True, dir_okay=True, exists=True), nargs=-1
)
@verbosity
@pass_config
def inventory_lint(
    config: Config, verbose: int, target: tuple[str], linter: tuple[str]
):
    """Lint YAML files in the provided paths.

    The command assumes that any YAML file found in the provided paths is part of a
    Commodore inventory structure."""
    config.update_verbosity(verbose)

    error_counts = []
    for t in target:
        lint_target = Path(t)
        for lint in linter:
            error_counts.append(LINTERS[lint](config, lint_target))

    errors = sum(error_counts)
    exit_status = 0 if errors == 0 else 1
    if errors == 0:
        click.secho("No errors 🎉 ✨", bold=True)
    if errors > 0:
        click.secho(f"Found {errors} errors", bold=True)

    sys.exit(exit_status)


@commodore.command(
    name="login",
    short_help="Login to Lieutenant",
)
@api_url_option
@oidc_discovery_url_option
@oidc_client_option
@pass_config
def commodore_login(
    config: Config, oidc_discovery_url: str, oidc_client: str, api_url: str
):
    """Login to Lieutenant"""
    config.oidc_client = oidc_client
    config.oidc_discovery_url = oidc_discovery_url
    config.api_url = api_url

    login(config)


@commodore.command(
    name="fetch-token",
    short_help="Fetch Lieutenant token",
)
@api_url_option
@oidc_discovery_url_option
@oidc_client_option
@pass_config
@verbosity
def commodore_fetch_token(
    config: Config,
    oidc_discovery_url: str,
    oidc_client: str,
    api_url: str,
    verbose: int,
):
    """Fetch Lieutenant token

    Get the token from the cache, if it's still valid, otherwise fetch a new token based
    on the provided OIDC config and Lieutenant URL.

    The command prints the token to stdout.
    """
    if api_url is None:
        raise click.ClickException(
            "Can't fetch Lieutenant token. Please provide the Lieutenant API URL."
        )

    config.api_url = api_url
    config.oidc_client = oidc_client
    config.oidc_discovery_url = oidc_discovery_url
    config.update_verbosity(verbose)

    token = fetch_token(config)
    click.echo(token)


def main():
    load_dotenv(dotenv_path=find_dotenv(usecwd=True))
    commodore.main(
        prog_name="commodore", auto_envvar_prefix="COMMODORE", max_content_width=100
    )
